from MacroMolecule import MacroMolecule
from Ligand import Ligand
from Hydration import Hydration
from WaterProbe import WaterProbe
from Void import Void
from ContactList import ContactList
from Point import Point
from UnitSurfacePoints import UnitSurfacePoints
from math import sqrt, pi
import sys
import random
import time
import re

# Using 'private' calls such as atom._x instead of atom.x for efficiency
# (bypasses the get_x function)

# Deal with the radii
import Radii
R = Radii.Radii()

class TmpWater(WaterProbe):
    """
    Container for temporary waters.
    """
    def __init__(self, atom_org_id, atom_org_name, atom_type, x, y, z,
                 sub_atom_type, molecule_instance, hydration_layer):
        self.point_clash = False
        self.point_polar_contacts = 0
        self.hydration_layer = hydration_layer
        WaterProbe.__init__(self, atom_org_id, atom_org_name, atom_type, x, y, z,
                 sub_atom_type, molecule_instance, 1, False)

def sorting_tmp_waters(something, other):
    """
    Sort the temporary waters based on the number of polar contacts.
    """
    return cmp(something.point_polar_contacts, other.point_polar_contacts)

def probe_sort(something, other):
    """
    Sort cavity pseudo atoms (probes) by size, largest last
    """
    return cmp( other._char_apol_radius, something._char_apol_radius )

def min_pol_rad_pointA(pointA, pointB, min_pol_radA):
    """
    Function for add_void_waters1.
    """
    distance_AB2 = pointA.get_distance2(pointB)
    min_pol_radA2 = min_pol_radA * min_pol_radA
    fraction = min_pol_radA / distance_AB2
    p = pointB - pointA
    p = p * fraction
    p = p + pointA
    return p

class TmpContainer(object):
    """
    Temporary container for add_void_waters1.
    """
    def __init__(self):
        self.polar_neighbor = None
        self.distance2 = None

def tmp_container_sort(something, other):
    """
    Sort temporary container by distance2.
    """
    return cmp(something.distance2, other.distance2)

def set_neighbors(atom_list):
    """
    Function for setting neighbors.
    
    atom_list: list

    This function sets the Atom object attributes:
    1) close_neighbors (list)
    2) far_neighbors (list)

    All the atom in the atom_list argument are considered as potential
    neighbors to each other.

    The close neighbor cutoff is defined as:
    atom1._probe_radius + atom2._probe_radius + 1

    The far neighbor cutoff is defined as:
    atom1._probe_radius + atom2._probe_radius + 5.3

    (The probe_radius is the maximum apolar radius)
    
    """
    for atom1 in atom_list:
        atom1.clear_neighbors()

    for a1Id, atom1 in enumerate(atom_list):
        a1Id = a1Id + 1
        for a2Id, atom2 in enumerate(atom_list):
            a2Id = a2Id + 1
            if a1Id <= a2Id:
                continue
            d2 = atom1.get_distance2(atom2)
            close_cutoff = ( atom1._probe_radius
                + atom2._probe_radius
                + 1.0 )
            far_cutoff = ( atom1._probe_radius
                + atom2._probe_radius
                + 5.3 )
            cc2 = close_cutoff * close_cutoff
            fc2 = far_cutoff * far_cutoff
            if d2 <= cc2:
                atom1.append_close_neighbor(atom2)
                atom1.insert_far_neighbor(atom2)        # Make sure that the
                                                        # close atoms are
                                                        # inserted before the
                                                        # far atoms for speed
                                                        # optimization of
                                                        # other steps
                atom2.append_close_neighbor(atom1)
                atom2.insert_far_neighbor(atom1)
            elif d2 <= fc2:
                atom1.append_far_neighbor(atom2)
                atom2.append_far_neighbor(atom1)
    
def update_neighbors(atom, atom_list):
    """
    Function for updating neighbors.
    
    atom: Atom
    atom_list: list
    
    This method is used to add a new atom to an existing set of
    atoms neighbor lists and to create the neighbor lists of the
    new atom. This is faster than the set_neighbors() method as
    the atom_list only has to be looped through once.

    See also set_neighbors().

    """
    atom.clear_neighbors()
    for atom1 in atom_list:
        if atom == atom1:
            continue
        d2 = atom.get_distance2(atom1)
        close_cutoff = ( atom._probe_radius
            + atom1._probe_radius
            + 1.0 )
        far_cutoff = ( atom._probe_radius
            + atom1._probe_radius
            + 5.3 )
        cc2 = close_cutoff * close_cutoff
        fc2 = far_cutoff * far_cutoff
        if d2 <= cc2:
            atom.append_close_neighbor(atom1)
            atom.insert_far_neighbor(atom1)     # Make sure that the close
                                                # are inserted before the far
                                                # atoms for speed optimization
                                                # of other steps
            atom1.append_close_neighbor(atom)
            atom1.insert_far_neighbor(atom)
        elif d2 <= fc2:
            atom.append_far_neighbor(atom1)
            atom1.append_far_neighbor(atom)

def get_res_id(atom):
    """
    Generate a unique residue id. For protein, ligands and water atoms.
    """
    s = 'res_id'
    if atom._sub_atom_type == 1:
        # protein atom
        # A_resid_resname_chain
        s = 'A_%s_%s_%s' % (atom._res_original_id, atom._res_type, atom._chain)
    if atom._sub_atom_type == 2:
        # ligand atom
        # B_9999_LIG_L
        s = 'B_9999_LIG_Z'
    if atom._sub_atom_type == 3:
        # water atom
        # C_resid_HOH_chain
        s = 'C_%s_HOH_%s' % (atom._res_original_id, atom._water_type)
    if atom._sub_atom_type == 4:
        # probe atom
        # D_resid_PPP_P
        s = 'D_%s_PPP_P' % atom._atom_org_id
    return s

AssemblyError = "assembly error"

class Assembly(object):
    """Class Assembly
    
    The class 'Assembly' has five attributes:
    1) macro_molecule - reference to a MacroMolecule object instance
    2) ligand         - reference to a Ligand object instance
    3) hydration      - reference to a Hydration object instance
    4) void           - reference to a Void object instance
    5) interaction    - reference to a ContactList instance

    Many of the methods in PRO_ACT2 depend upon the specific environment of an
    atom. The environment mainly consists of neighboring atoms. There are
    several types of 'things' classified as atoms:
    1) MacroMoleculeAtom instances stored in the container MacroMolecule
    2) LigandAtom instances stored in the container Ligand
    3) WaterProbe instances stored in the container Hydration
    4) VoidProbe instances stored in the container Void

    Examples:
    >>> import Assembly
    >>> import Radii
    >>> R = Radii.Radii()
    >>> R.radii_set = 'LI_AND_NUSSINOV'
    >>> A = Assembly.Assembly()
    >>>
    >>> A.macro_molecule.atom_input(2, 'GLN', 'A', 17, ' N  ', 'N', 30.607, 37.334, 3.305, 1, 1)
    >>> A.macro_molecule.atom_input(2, 'GLN', 'A', 18, ' CA ', 'C', 30.158, 36.492, 2.199, 1, 1)
    >>> A.macro_molecule.atom_input(2, 'GLN', 'A', 19, ' C  ', 'C', 30.298, 35.041, 2.643, 1, 1)
    >>> A.macro_molecule.atom_input(2, 'GLN', 'A', 20, ' O  ', 'O', 31.401, 34.494, 2.763, 1, 1)
    >>>
    >>> Assembly.set_neighbors(A.macro_molecule)
    >>>
    >>> import UnitSurfacePoints
    >>> u  = UnitSurfacePoints.UnitSurfacePoints()
    >>> for atom in A.macro_molecule:
    ...     atom.set_surface_points(u, atom.close_radius, 0, 0)
    ...     atom.set_accessibility(u)
    >>>
    >>> environment = []
    >>> next_hoh_id = A.add_water(u, A.macro_molecule, environment, 1, 1001)
    >>> A.set_water_type(A.hydration, u, 1)
    >>> A.correct_water_maps(A.hydration, A.macro_molecule)
    >>>
    >>> A.probe_fill(u, A.macro_molecule)
    >>>
    >>> A.set_contacts(A.macro_molecule, A.hydration)
    >>> A.interaction.calculate_contact_counts()

    """
    
    def __init__(self):
        """
        Initialise the object.

        macro_molecule: MacroMolecule
        ligand: Ligand
        hydration: Hydration
        void: Void
        interaction: ContactList
        
        """
        self._macro_molecule = MacroMolecule()
        self._ligand = Ligand()
        self._hydration = Hydration()
        self._void = Void()
        self._interaction = ContactList()
    
    @property
    def macro_molecule(self):
        """
        Get macro_molecule attribute.
        """
        return self._macro_molecule
    
    @property
    def ligand(self):
        """
        Get the ligand attribute
        """
        return self._ligand
    
    @property
    def hydration(self):
        """
        Return the hydratoin attribute.
        """
        return self._hydration
    
    @property
    def void(self):
        """
        Return the void attribute.
        """
        return self._void
    
    @property
    def interaction(self):
        """
        Return the interaction attribute.
        """
        return self._interaction
    
    def add_water(self, unit_points, atom_list, environment_list,
        hydration_layer, res_start_id, only_interaction=False):
        """function add_water
        
        unit_points: UnitSurfacePoints
        atom_list: list
        environment_list: list
        hydration_layer: int
        res_start_id: int
        only_interaction: boolean

        This function adds waters to the accessible surfaces of the
        atoms in atom_list.

        The unit_points argument is required to set the accessiblity
        of the added waters.

        The hydration_layer argument specifies which hydration layer is being
        added. If the hydration layer is 1 waters are only added to
        MacroMoleculeAtoms and LigandAtoms, any surfaces of waters (including
        experimental waters) are ignored, but are considered in terms of
        clashes if they are part of the MacroMolecule-, Ligand- Atoms'
        neighbor lists. If the hydration layer is greater than 1 waters are
        only added to WaterProbe atoms, any surfaces of MacroMolecule-
        Ligand-Atoms are ignored, but are considered in terms of clashes.

        If no hydration layer is smaller than 1 an AssemblyError is raised.

        This is an improved algorithm of water placement compared to that used
        in the original pro_act.
        1) Generates random atom list
        2) Creates surface point list from random atom list (where there is
        accessible surface area and no steric clashes)
        3) Counts potential polar contacts at each surface point
        4) Sorts the surface points by number of potential polar contacts
        5) Loops through the surface points in order of potential polar
        contacts. Adds a water. Removes any surface points that clashes with
        the added water.
        
        """
        # Set variables
        hydration = self.hydration
        cutoff = R.water_radius * 2
        atoms_ordered = []
        atoms_random = []
        dynamic_neighbor_list = atom_list[:]
        dynamic_neighbor_list += environment_list[:]

        # Create atom list
        if hydration_layer == 1:     # first hydration layer add waters to 
            for atom in atom_list:   # macromolecule and ligand atoms
                if only_interaction:
                    if atom._include != 1:
                        continue
                if ( atom._sub_atom_type == 3
                    or atom._sub_atom_type == 4 ):
                    continue
                atoms_ordered.append(atom)
        elif hydration_layer > 1:
            tmp_waters = []
            for water in hydration:
                tmp_waters.append(water)
            atoms_ordered = tmp_waters[:]
        else:
            raise AssemblyError, "add_water(), requires hydration_layer 1+"

        # Randomise atom list
        while atoms_ordered:
            atom = random.choice(atoms_ordered)
            atoms_ordered.remove(atom)
            atoms_random.append(atom)
        
        tmp_water_list = []
        
        # Store temporary waters in container and count the number of
        # potential polar contacts
        for atom in atoms_random:
            surface = atom._surface
            for surface_point in surface._surface_point_list:

                if not( surface_point._accessibility == 0
                    or surface_point._accessibility == 2):
                    continue
                tmp_water_list.append(TmpWater(
                    None,
                    ' O  ',
                    'O',
                    surface_point._x,
                    surface_point._y,
                    surface_point._z,
                    3,
                    1,
                    hydration_layer))
                polar_contact_count = 0
                for atom2 in atom._far_neighbors:
                    if tmp_water_list[-1].in_contact(atom2) == 0:
                        polar_contact_count = polar_contact_count + 1
                tmp_water_list[-1].point_polar_contacts = polar_contact_count

        tmp_water_list.sort(sorting_tmp_waters)
        tmp_water_list.reverse()

        # Add waters
        while len(tmp_water_list) != 0:
            tmp_water = tmp_water_list[0]
            self._hydration.atom_input(
                hydration_layer,
                res_start_id,
                None,
                ' O  ',
                'O',
                tmp_water._x,
                tmp_water._y,
                tmp_water._z,
                3,
                1)
            update_neighbors(
                self._hydration[-1],
                dynamic_neighbor_list)
            dynamic_neighbor_list.append(self._hydration[-1])
            res_start_id += 1
            to_do = len(tmp_water_list)
            while to_do != 1:
                to_do = to_do - 1
                tmp_water2 = tmp_water_list[to_do]
                if tmp_water.clashes(tmp_water2, cutoff):
                    tmp_water_list.remove(tmp_water2)
            tmp_water_list.remove(tmp_water)

        # Set accessibility of all waters new waters (this implies that the
        # accessibility of previously added waters will remain the same)
        for water_id, water in enumerate(hydration):
            water_id = water_id + 1
            if water.hydration_layer == hydration_layer:
                water.set_surface_points(unit_points, water._close_radius, 0, 0)
                water.set_accessibility(unit_points)
        
        return res_start_id
        
    def add_water_original(self, unit_points, atom_list, environment_list,
        hydration_layer, res_start_id, only_interaction=False):
        """function add_water_original
        
        unit_points: UnitSurfacePoints
        atom_list: list
        hydration_layer: int
        res_start_id: int
        only_interaction: boolean

        This function adds waters to the accessible surfaces of the
        atoms in atom_list.

        The unit_points argument is required to set the accessiblity
        of the added waters.

        The hydration_layer argument specifies which hydration layer is
        being added. If the hydration layer is 1 waters are only
        added to MacroMoleculeAtoms and LigandAtoms, any surfaces of waters
        (including experimental waters) are ignored, but are considered
        in terms of clashes if they are part of the MacroMolecule-, Ligad-
        Atoms' neighbor lists. If the hydration layer is greater than 1
        waters are only added to WaterProbe atoms, any surfaces of
        MacroMolecule- Ligand-Atoms are ignored, but are considered in terms
        of clashes.

        If no hydration layer is smaller than 1 an AssemblyError is raised.
        
        This mehtod is based on the original pro_act algorithm of adding
        waters randomly to polar surface points first.
        
        """
        # Set variables
        hydration = self.hydration
        cutoff = R.water_radius * 2
        polar_atoms_ordered = []
        apolar_atoms_ordered = []
        atoms_random = []
        dynamic_neighbor_list = atom_list[:]
        dynamic_neighbor_list += environment_list[:]

        # Create atom list
        if hydration_layer == 1:     # first hydration layer add waters to 
            for atom in atom_list:   # macromolecule and ligand atoms
                if only_interaction:
                    if atom._include != 1:
                        continue
                if ( atom._sub_atom_type == 3
                    or atom._sub_atom_type == 4 ):
                    continue
                if atom._atom_polarity == 2:  # apolar
                    apolar_atoms_ordered.append(atom)
                else:                        # polar
                    polar_atoms_ordered.append(atom)
        elif hydration_layer > 1:
            tmp_waters = []
            for water in hydration:
                tmp_waters.append(water)
            polar_atoms_ordered = tmp_waters[:]
        else:
            raise AssemblyError, "add_water(), requires hydration_layer 1+"

        # Randomise atom list
        while polar_atoms_ordered:
            atom = random.choice(polar_atoms_ordered)
            polar_atoms_ordered.remove(atom)
            atoms_random.append(atom)
        while apolar_atoms_ordered:
            atom = random.choice(apolar_atoms_ordered)
            apolar_atoms_ordered.remove(atom)
            atoms_random.append(atom)
        
        for atom in atoms_random:
            surface = atom._surface
            for surface_point in surface._surface_point_list:

                if not( surface_point._accessibility == 0
                    or surface_point._accessibility == 2):
                    continue
                clash = False
                for neighbor in atom._far_neighbors:
                    if surface_point.clashes(neighbor, cutoff):
                        clash = True
                        break
                if not(clash):
                    self._hydration.atom_input(
                        hydration_layer,
                        res_start_id,
                        None,
                        ' O  ',
                        'O',
                        surface_point._x,
                        surface_point._y,
                        surface_point._z,
                        3,
                        1)
                    update_neighbors(
                        self._hydration[-1],
                        dynamic_neighbor_list)
                    dynamic_neighbor_list.append(self._hydration[-1])
                    res_start_id += 1
        
        # Set accessibility of all waters new waters (this implies that the
        # accessibility of previously added waters will remain the same)
        for water_id, water in enumerate(hydration):
            water_id = water_id + 1
            if water.hydration_layer == hydration_layer:
                water.set_surface_points(unit_points, water._close_radius, 0, 0)
                water.set_accessibility(unit_points)
        
        return res_start_id

    def add_water_random(self, unit_points, atom_list, environment_list,
        hydration_layer, res_start_id, only_interaction=False):
        """function add_water_random
        
        unit_points: UnitSurfacePoints
        atom_list: list
        hydration_layer: int
        res_start_id: int
        only_interaction: boolean

        This function adds waters to the accessible surfaces of the
        atoms in atom_list.

        The unit_points argument is required to set the accessiblity
        of the added waters.

        The hydration_layer argument specifies which hydration layer is
        being added. If the hydration layer is 1 waters are only
        added to MacroMoleculeAtoms and LigandAtoms, any surfaces of waters
        (including experimental waters) are ignored, but are considered
        in terms of clashes if they are part of the MacroMolecule-, Ligad-
        Atoms' neighbor lists. If the hydration layer is greater than 1
        waters are only added to WaterProbe atoms, any surfaces of
        MacroMolecule- Ligand-Atoms are ignored, but are considered in terms
        of clashes.

        If no hydration layer is smaller than 1 an AssemblyError is raised.

        This algorithm is intended to be used as a control against which the
        other add_water methods can be validated. This method simply places
        waters randomly on the protein/ligand/complex surface.
        
        """
        # Set variables
        hydration = self.hydration
        cutoff = R.water_radius * 2
        atoms_ordered = []
        atoms_random = []
        dynamic_neighbor_list = atom_list[:]
        dynamic_neighbor_list += environment_list[:]

        # Create atom list
        if hydration_layer == 1:     # first hydration layer add waters to 
            for atom in atom_list:   # macromolecule and ligand atoms
                if only_interaction:
                    if atom._include != 1:
                        continue
                if ( atom._sub_atom_type == 3
                    or atom._sub_atom_type == 4 ):
                    continue
                atoms_ordered.append(atom)
        elif hydration_layer > 1:
            tmp_waters = []
            for water in hydration:
                tmp_waters.append(water)
            atoms_ordered = tmp_waters[:]
        else:
            raise AssemblyError, "add_water(), requires hydration_layer 1+"

        # Randomise atom list
        while atoms_ordered:
            atom = random.choice(atoms_ordered)
            atoms_ordered.remove(atom)
            atoms_random.append(atom)
        
        for atom in atoms_random:
            surface = atom._surface
            for surface_point in surface._surface_point_list:

                if not( surface_point._accessibility == 0
                    or surface_point._accessibility == 2):
                    continue
                clash = False
                for neighbor in atom._far_neighbors:
                    if surface_point.clashes(neighbor, cutoff):
                        clash = True
                        break
                if not(clash):
                    self._hydration.atom_input(
                        hydration_layer,
                        res_start_id,
                        None,
                        ' O  ',
                        'O',
                        surface_point._x,
                        surface_point._y,
                        surface_point._z,
                        3,
                        1)
                    update_neighbors(
                        self._hydration[-1],
                        dynamic_neighbor_list)
                    dynamic_neighbor_list.append(self._hydration[-1])
                    res_start_id += 1
        
        # Set accessibility of all waters new waters (this implies that the
        # accessibility of previously added waters will remain the same)
        for water_id, water in enumerate(hydration):
            water_id = water_id + 1
            if water.hydration_layer == hydration_layer:
                water.set_surface_points(unit_points, water._close_radius, 0, 0)
                water.set_accessibility(unit_points)

        return res_start_id
        
    def set_water_type_original(self, water_list, hydration_layers):
        """function set_water_type_original
        
        water_list: list
        hydration_layers: int  (number of hydration layers)

        This method sets the water types. The water types are classified as:
        1) Buried  -- (=0)
        2) Surface -- (=1)
        3) Cleft   -- (>1)
        4) Further -- (<0)
        The further type will only occur if more than 1 hydration layer is
        added and the negative value refers to the distance from the surface
        layer, in other words -1 is of type further and is one layer out from
        the surface layer.
        
        """
        # For every water atom draw a dot surface at the radius of water
        number_of_buried_waters = 0
        for water in water_list:
            surface = water._surface
            for surface_point in surface._surface_point_list:
                surface_point.update_surface_point(
                    water._close_radius,
                    water._x,
                    water._y,
                    water._z,
                    0, 0)
            water.water_type = 0
            number_of_buried_waters = number_of_buried_waters + 1

        npass = 0     # passes through the while loop, referes to 'buridness'
        ncount = 1    # condition for continuing while loop
        while ncount > 0:
#            print "ncount:", ncount
            ncount = 0                   # is this going to be the last pass?
            npass = npass + 1            # next degree of buriedness
            wtype = npass - hydration_layers  # buridness 'layer' is corrected
            if wtype > -1:                   # for the number of hydration
                wtype = wtype + 1            # layers, also make sure that
                                             # '0' is reserved for buried
                                             # waters

            for water_id, water in enumerate(water_list):
                if water._water_type != 0:
                    continue
                water_id = water_id + 1
#                print "water", water_id,
                surface = water._surface
                solvent_accessible = False
                for point_id, surface_point in enumerate(surface._surface_point_list):
                    next_surface_point = False
                    point_solvent_accessible = True
                    point_id = point_id + 1
#                    print "point", point_id
                    for neighbor_id, neighbor in enumerate(water._far_neighbors):
                        neighbor_id = neighbor_id + 1
                        neighbor_clash = False
                        if ( neighbor._sub_atom_type == 3 # water
                            and neighbor._water_type != 0):
                            continue
                        cutoff = neighbor._far_radius  # default to apolar
                        if neighbor._atom_polarity == 0:  # polar
                            cutoff = neighbor._close_radius
                        if surface_point.clashes(neighbor, cutoff):
                            # Not solvent accessible
                            neighbor_clash = True
                            point_solvent_accessible = False
                            next_surface_point = True
#                            print "clashes with neighbor", neighbor_id,
                            break  # out of neighbor loop
                                   # straight to next surface point
                                   # resets clash to False
                        else:
                            pass
#                            print "doesn't clash with neighbor", neighbor_id

                    if point_solvent_accessible:
#                        print "point solvent accessible",
                        solvent_accessible = True

                    if next_surface_point:
#                        print "next surface point..."
                        continue  # next surface point
                    
                    if solvent_accessible:
#                        print "break out of surface point...",
                        break  # out of surface point
                        
                if solvent_accessible:
                    water.water_type = wtype
                    ncount = ncount + 1
#                    print "solvent accessible! pass", npass, "ncount", ncount
                else:
                    pass
#                    print "not solvent accessible! pass", npass, "ncount", ncount

#        print "ncount", ncount, "end"

    def set_water_type(self, water_list, unit_points, hydration_layers):
        """function set_water_type
        
        water_list: list
        unit_points: reference to unitSurfacePoints
        hydration_layers: int  (number of hydration layers)

        This algorithm is changed from the original algorithm in that the
        after the first pass of determining the solvent accessibilities of the
        waters the water types are determined in terms of polar contacts to
        other waters. The cutoff used is maximum apolar radius * 2. Using the
        maximum apolar radius is not in line with the original paper that
        states that it bases its water types on hydrogen bond networks.
        However, that leads to buried waters on the surface as the water
        placement is not perfect.

        Also in order to deal with multiple hydration layers of small
        molecules where the second solvation shell does not prevent the first
        hydration shell waters from having solvent accessible surfaces, the
        first step of the algorithm only looks for solvent accessible surfaces
        of waters added in the outmost hydration layer.

        This algorithm works well for up to two hydration layers, after that
        it is not great.

        This method sets the water types. The water types are classified as:
        1) Buried  -- (=0)
        2) Surface -- (=1)
        3) Cleft   -- (>1)
        4) Further -- (<0)
        The further type will only occur if more than 1 hydration layer is
        added and the negative value refers to the distance from the surface
        layer, in other words -1 is of type further and is one layer out from
        the surface layer.
        
        """
        
        # For every water atom draw a dot surface at the radius of water
        number_of_buried_waters = 0
        for water in water_list:
            try:
                water.clear_surface()
            except AttributeError:
                pass
            water.set_surface_points(unit_points, water._close_radius, 0, 0)
            water.water_type = 0
            number_of_buried_waters = number_of_buried_waters + 1

        # Find the waters that have solvent accessible surfaces
        npass = 1
        wtype = npass - hydration_layers  # buridness 'layer' is corrected
        if wtype > -1:                   # for the number of hydration
            wtype = wtype + 1            # layers, also make sure that
                                         # '0' is reserved for buried
                                         # waters
        for water in water_list:
            if water._water_type != 0:
                continue
            if water.hydration_layer != hydration_layers:
                continue
            surface = water._surface
            solvent_accessible = False
            point_id = 0
            for surface_point in surface._surface_point_list:
                next_surface_point = False
                point_solvent_accessible = True
                point_id = point_id + 1
#                    print "point", point_id
                neighbor_id = 0
                for neighbor in water._far_neighbors:
                    neighbor_id = neighbor_id + 1
                    neighbor_clash = False
                    if ( neighbor._sub_atom_type == 3   # water
                        and neighbor._water_type != 0):
                        continue
                    if neighbor._sub_atom_type == 4:    # probe
                        continue
                    cutoff = neighbor._far_radius  # default to apolar
                    if neighbor._atom_polarity == 0:  # polar
                        cutoff = neighbor._close_radius

                    if surface_point.clashes(neighbor, cutoff):
                        # Not solvent accessible
                        neighbor_clash = True
                        point_solvent_accessible = False
                        next_surface_point = True
#                            print "clashes with neighbor", neighbor_id,
                        break  # out of neighbor loop
                               # straight to next surface point
                               # resets clash to False

                if point_solvent_accessible:
#                        print "point solvent accessible",
                    solvent_accessible = True

                if next_surface_point:
#                        print "next surface point..."
                    continue  # next surface point
                
                if solvent_accessible:
#                        print "break out of surface point...",
                    break  # out of surface point

            # Set water type to 1 if solvent accessible
            if solvent_accessible:
                water.water_type = wtype
                    
        # Determine the waters in contact with waters of previous layers
        ncount = 1    # condition for continuing while loop
        while ncount > 0:
#            print "ncount:", ncount
            ncount = 0                   # is this going to be the last pass?
            npass = npass + 1            # next degree of buriedness
            wtype = npass - hydration_layers  # buridness 'layer' is corrected
            if wtype > -1:                   # for the number of hydration
                wtype = wtype + 1            # layers, also make sure that
                                             # '0' is reserved for buried
                                             # waters

            tmp_surface_water = []
            for water1 in water_list:
                if water1._water_type == 0:
                    continue
                tmp_surface_water.append(water1)
            for water1 in tmp_surface_water:
                for water2 in water_list:
                    if water2._water_type != 0:
                        continue
                    if water1.clashes(water2, R.water_apolar_radius*2):
                        water2.water_type = wtype
                        ncount = ncount + 1
    
    def correct_water_maps(self, water_list, atom_list):
        """function correct_water_maps
        
        water_list: list
        atom_list: list

        This method first sets the accessibility of surface accessible
        points close to buried waters to probe accessible. It then corrects
        all probe accessible points on the basis of:
        1) Clashes with non-buried waters (accessiblity set to -2)
        2) Clashes with neighboring atoms solvent accessible surface
           (accessibility set to -3)
        3) Clashes with atoms own solvent accessible surface (accessibility
           set to -3)

        (In the probe_fill method an accessiblity of -3 is still considered
        probe accessible.)
        
        """
        water_radius2 = R.water_radius*2
        water_apolar_radius = R.water_apolar_radius
        # Regard buried solvent sites as probe accessible not solvent
        # accessible this requires ammendments of the access maps for
        # the buried waters neighboring atoms
        for water in water_list:
            if water._water_type != 0:   # only buried waters
                continue

            for neighbor in water._far_neighbors:
                if ( neighbor._sub_atom_type == 1       # only macromol
                    or neighbor._sub_atom_type == 2 ):  # and ligand
                    pass
                else:
                    continue
                if neighbor._include != 1:      # only use included atoms
                    continue
                surface = neighbor._surface
                for surface_point in surface._surface_point_list:
                    if ( surface_point._accessibility == 0
                        or surface_point._accessibility == 2 ): 
                        # only polar or apolar accessible
                        pass
                    else:
                        continue

                    if surface_point.clashes(water, water_radius2):
                        surface_point.update_surface_point(
                            neighbor._probe_radius,
                            neighbor._x, neighbor._y, neighbor._z,
                            -1, 0)
        
        # Alter point accessiblity of atoms in atom_list
        for atom in atom_list:

            if atom._include != 1:   # only use included atoms
                continue
            surface = atom._surface

            # Check for probe clashes with potential inserted waters
            
            # Loop through all the surface points
            for surface_point in surface._surface_point_list:
                sp_changed = False  # surface point changed

                # Only check probe accessible points
                if surface_point._accessibility != -1:
                    continue

                # Loop through the atoms neighbors
                for neighbor in atom._far_neighbors:

                    # If the neighbor is a water that is not buried
                    if (neighbor._sub_atom_type == 3      # water
                        and neighbor._water_type != 0):  # not buried

                        # Check for probe contact with neighboring water
                        if surface_point.clashes(neighbor, water_apolar_radius):
                            # Clash! Alter accessibility...
                            surface_point.set_accessibility(-2)
                            sp_changed = True
                            break
                

                    # If the neighbor is an macromolecule or ligand atom
                    if (neighbor._sub_atom_type == 1       # macromolecule
                        or neighbor._sub_atom_type == 2):  # ligand
                    
                        if neighbor._include != 1:  # only use included
                            continue                    # atoms
                        # Check for clashes with neighboring atoms
                        # solvent surfaces
                        n_surface = neighbor._surface

                        # The neighbor has solvent accessible surface
                        # loop through the individual surface points
                        for n_surface_point in n_surface._surface_point_list:
                            # Next unless the neighbor surface point is
                            # polar or apolar solvent accessible
                            if ( n_surface_point._accessibility == 0
                                or n_surface_point._accessibility == 2 ):
                                pass
                            else:
                                continue
                            
                            if (surface_point._accessibility == -1):
                                pass
                            else:
                                continue

                            # The n_surface_point is polar or apolar solvent
                            # accessible
                            # Check from contact with atom surface_point
                            if surface_point.clashes(n_surface_point, water_apolar_radius):
                                # Clash! Alter accessibility...
                                surface_point.set_accessibility(-3)
                                sp_changed = True
                                break

                if sp_changed:
                    continue

                # Finally check for point clashes with own surface

                # There is solvent accessible surface to check against
                # loop through the surface points a second time
                for surface_point2 in surface._surface_point_list:

                    # If polar or apolar solvent accessible
                    if ( surface_point2._accessibility == 0
                        or surface_point2._accessibility == 2):
                        # The second surface point is polar or apolar
                        # accessible
                        # Check for contact with atom surface_point
                        if surface_point._accessibility == -1:
                            if surface_point.clashes(surface_point2, water_apolar_radius):
                                # Clash! Alter accessibility...
                                surface_point.set_accessibility(-3)
                                break

            # Update the atom's solvent and probe accessible counts
            surface.sa_point_count
            surface.pa_point_count
    
    def probe_fill(self, unit_points, atom_list, probe_min_cutoff=0.8):
        """function probe_fill

        unit_points: UnitSurfacePoints
        atom_list: list

        """
        # Insert probes on the apolar probe accessible surface of the protein
        # (or ligand) atoms and allow them to move and expand to fill the
        # voids in the protein(-ligand) structure

        # If the onlyBindingSite or onlyInterface option has been set all the
        # neighbors might not have surface points and accessibilities
        for atom in atom_list:
            for neighbor in atom._far_neighbors:
                try:
                    neighbor._surface
                except AttributeError:
                    # The surface is not set so set it
                    neighbor.set_surface_points(unit_points,
                        neighbor._close_radius, 0, 0)
                    neighbor.set_accessibility(unit_points)
        

        void = self.void
                
        # loop forever until the minimum radius of the probe is less or eqaul
        # to probe_min_cutoff
        npass = 0
        radius_probe_min = R.water_radius
        while radius_probe_min >= probe_min_cutoff:      # while loop N
            npass = npass + 1

            # loop through all the atoms of the protein (and the ligand)
            for atom in atom_list:
                # set the radius of the probe to the greater of
                # the atom's probe radius (=maximum apolar radius) or
                # the atom's characteristic apolar radius + the minimum probe
                radius_probe = max ( atom._probe_radius,
                    atom._char_apol_radius+radius_probe_min)
                
                # loop through all the surface points of the atom
                surface = atom._surface
                for surface_point in surface._surface_point_list:

                    if surface_point._flag != 1:
                        pass
                    else:
                        continue  # next surface point

                    # check if surface point is probe accessible (-1 or -3)
                    if ( surface_point._accessibility == -1 
                         or surface_point._accessibility == -3 ):
                        pass
                    else:
                        continue  # next surface point
                    
                    # set some variables
                    unit = surface_point._unit_surface_point
                    probe_x = unit._x*radius_probe + atom._x
                    probe_y = unit._y*radius_probe + atom._y
                    probe_z = unit._z*radius_probe + atom._z
                    distance_min1 = ( radius_probe
                        -atom._char_apol_radius
                        -0.0001 )
                    scale = 0.2
                    step_x = 0.0
                    step_y = 0.0
                    step_z = 0.0
                    
                    # loop forever until the scale is 0.025 or smaller
                    mpass = 0

                    while scale > 0.025:     # while loop M
                        mpass = mpass + 1
                        next_pass_through_m = False
                        break_out_of_m = False
                        next_point = False

                        # move the probe
                        tmp_x = probe_x + scale*step_x
                        tmp_y = probe_y + scale*step_y
                        tmp_z = probe_z + scale*step_z
                        tmpProbe = Point(tmp_x, tmp_y, tmp_z)
                        
                        # set some more varibles
                        distance = sqrt( tmpProbe.get_distance2(atom) )
                        distance_min2 = distance - atom._char_apol_radius
                        distance_min3 = 0.0
                        kmin = atom

                        # check for clash with atom's probe radius
                        if distance <= atom._probe_radius - 0.001:
                            distance_min3 = 1.0
                        if ( distance_min2 < distance_min1
                            or distance_min3 == 1.0 ):
                            if mpass == 1:
                                next_point = True
                                break  # out of M
                            else:
                                scale = scale*0.5
                                if scale <= 0.025:
                                    break  # out of M
                                else:
                                    continue
                        else:
                            pass

                        # check for clash with neighbors' probe radius
                        # buried water neighbores are ignored
                        for neighbor in atom._far_neighbors:
                    
                            # ignore buried waters
                            if neighbor._sub_atom_type == 3:
                                if neighbor._water_type == 0:
                                    continue
                                    pass
                                else:
                                    pass

                            # check for clash with neighbors characteristic
                            # apolar radius, if clash alter the distance_min2
                            # and set kmin to the neighbor id
                            # this will be used to move the probe
                            nDistance = sqrt( tmpProbe.get_distance2(neighbor) )
                            if neighbor._sub_atom_type == 4:
                                nDistance = nDistance + min(
                                    neighbor._char_apol_radius,
                                    distance_min2)
                            if ((nDistance - neighbor._char_apol_radius)
                                < distance_min2):
                                distance_min2 = ( nDistance 
                                    - neighbor._char_apol_radius )
                                kmin = neighbor

                            # check for clash with neighbors' probeRadius
                            if nDistance <= neighbor._probe_radius - 0.001:
                                distance_min3 = 1.0
                            if ( distance_min2 < distance_min1
                                or distance_min3 == 1.0):
                                if mpass == 1:
                                    next_point = True
                                    break_out_of_m = True
                                    break  # out of neighbor loop
                                else:
                                    scale = scale*0.5
                                    if scale <= 0.025:
                                        break_out_of_m = True
                                        break  # out of neighbor loop
                                    else:
                                        next_pass_through_m = True
                                        break  # out of neighbor loop
                            else:
                                pass

                        if next_pass_through_m:
                            continue
                        if break_out_of_m:
                            break
                            
                        # check for clashes with atom's own solvent
                        # accessible surface
                        tmp_tumin = 1000.0   # random large number
                        surface2 = atom._surface
                        for sPoint2 in surface2._surface_point_list:

                            if ( (sPoint2._accessibility == 0
                                or sPoint2._accessibility == 2)
                                and sPoint2._flag == 0):
                                sDist2 = tmpProbe.get_distance2(sPoint2)
                                if sDist2 <= tmp_tumin*tmp_tumin:
                                    tmp_tumin = sqrt( sDist2 )
                                    if tmp_tumin < R.water_apolar_radius:
                                        if mpass == 1:
                                            next_point = True
                                            break_out_of_m = True
                                            break  # out of surface point
                                        else:
                                            scale = scale*0.5
                                            if scale <= 0.025:
                                                break_out_of_m = True
                                                break  # out of surface point
                                            else:
                                                next_pass_through_m = True
                                                break  # out of surface point
                                    else:
                                        pass
                            else:
                                pass

                        if next_pass_through_m:
                            continue
                        if break_out_of_m:
                            break  # out of M

                        # check for clashes with neighbors' solvent
                        # accessible surface
                        for neighbor in atom._far_neighbors:
                            
                            if ( neighbor._sub_atom_type == 1
                                or neighbor._sub_atom_type == 2 ):
                                # neighbor is macromolecule or ligand
                                pass
                            else:
                                continue

                            n_surface = neighbor._surface
                            if n_surface.sa_point_count == 0:
                                continue
                            else:
                                pass
                            for nPoint in n_surface._surface_point_list:
                                if ( nPoint._accessibility == 0
                                    or nPoint._accessibility == 2 ):
                                    nDist2 = tmpProbe.get_distance2(nPoint)
                                    if nDist2 <= tmp_tumin*tmp_tumin:
                                        tmp_tumin = sqrt( nDist2 )
                                        if tmp_tumin < R.water_apolar_radius:
                                            if mpass == 1:
                                                next_point = True
                                                break_out_of_m = True
                                                break  # out of nPoint
                                            else:
                                                scale = scale*0.5
                                                if scale <= 0.025:
                                                    break_out_of_m = True
                                                    break  # out of nPoint
                                                else:
                                                    next_pass_through_m = True
                                                    break  # out of nPoint
                                        else:
                                            pass
                                    else:
                                        pass
                                else:
                                    pass
                            if next_pass_through_m or break_out_of_m:
                                break  # out of neighbor
                        if next_pass_through_m:
                            continue
                        if break_out_of_m:
                            break
                            
                        # All checks passed
                        # Calculate next move and
                        # (re)set some variables
                        probe_x = tmpProbe._x
                        probe_y = tmpProbe._y
                        probe_z = tmpProbe._z
                        distance_min1 = distance_min2
                        scale = 0.2
                        
                        if kmin != atom:
                            step_x = ( ( probe_x - kmin._x )
                                / sqrt(tmpProbe.get_distance2(kmin)))
                            step_y = ( ( probe_y - kmin._y )
                                / sqrt(tmpProbe.get_distance2(kmin)))
                            step_z = ( ( probe_z - kmin._z )
                                / sqrt(tmpProbe.get_distance2(kmin)))
                        else:
                            step_x = ( ( probe_x - atom._x )
                                / sqrt(tmpProbe.get_distance2(atom)))
                            step_y = ( ( probe_y - atom._y )
                                / sqrt(tmpProbe.get_distance2(atom)))
                            step_z = ( ( probe_z - atom._z )
                                / sqrt(tmpProbe.get_distance2(atom)))
                        
                    # Check if the probe insertion has failed any tests
                    if (next_point):
                        continue

                    # The probe insertion has convereged
                    # Add to void_probe_list

                    void.atom_input(
                        None,
                        'Probe',
                        'Probe',
                        probe_x, probe_y, probe_z,
                        4,
                        0)
                    probe_list = void
                    new_probe = probe_list[-1]

                    # Make sure the probe radius is smaller (or equal to)
                    # the charachteristic apolar radius of water
                    # and set the probe's surface points
                    new_probe_radius = 0.0
                    if distance_min1 > R.water_apolar_radius:
                        new_probe_radius = R.water_apolar_radius
                    else:
                        new_probe_radius = distance_min1
                    new_probe._char_apol_radius = new_probe_radius
                    new_probe.probe_radius = new_probe_radius

                    # Update neighbor lists
                    updated_neighbor_list = atom_list + probe_list
                    update_neighbors(new_probe, updated_neighbor_list)
                    
            radius_probe_min = radius_probe_min - 0.2  # to get out of while loop

        self.void.sort(probe_sort)
        for i, p in enumerate(self.void):
            p._atom_org_id = i + 1 

    def add_void_waters1(self, res_start_id, probe_list, all_non_probe_atoms):
        """function add_void_waters1

        res_start_id: int
        probe_list: list
        all_non_probe_atoms: list

        This is an experimental function which tries to add waters to
        VoidProbes.
        1) The VoidProbe has at least two polar atoms that are close_neighbors.
        2) Get all the VoidProbes polar atoms that are close_neighbors.
        3) Get the midpoints of the lines between the polar atoms minimum
        polar radius.
        4) The average coordinates of the midpoints is where the water will be
        placed.
        5) Check for clashes VoidProbes farNeighbors using:
            i)  minimum polar radius + minimum polar water radius for polar atoms
            ii) apolar radius + minimum polar water radius for apolar atoms
        6) If no clashes place water and update neighbors.

        Point A                    Point B
        |              Min Pol Rad of B  |
        |  Min Pol Rad of A        |     |
        |   |                      |     |
        A--mpr--------------------mpr----B
        
        To find Min Pol Rad of A:
        MinPolRadAPoint = A + ( min_pol_radA / distance(AB) )*(B - A)
        
        """
        dynamic_neighbor_list = all_non_probe_atoms[:]
        min_water_radius = R.water_radius - ( R.water_radius_max - R.water_radius )
        
        for probe in probe_list:
            tmp_list = []
            for neighbor in probe._close_neighbors:
                if neighbor._sub_atom_type != 4:
                    if neighbor._atom_polarity == 0:
                        tmp_list.append(TmpContainer())
                        tmp_list[-1].polar_neighbor = neighbor
                        tmp_list[-1].distance2 = probe.get_distance2(neighbor)
    #                   print "@balllist {add_void_waters1} colour=red radius=0.1"
    #                   print "{ }",
    #                   print round(neighbor._x, 3),
    #                   print round(neighbor._y, 3),
    #                   print round(neighbor._z, 3)
            if len(tmp_list) < 2:
                # Not enough polar neighbors to justify placing water
                continue
            tmp_list.sort(tmp_container_sort)
            polar_neighbors = []
            i = 0
            for neighbor in tmp_list:
                if i < 4:
                    # A water can only form 4 h-bonds!
                    # Choose the closest four polar neighbors
                    polar_neighbors.append(neighbor.polar_neighbor)
                    i = i + 1
                else:
                    break
                
            midpoints = []
            for polar_atom1 in polar_neighbors:
#               print "@balllist {add_void_waters1} colour=blue radius=0.1"
#               print "{ }",
#               print round(polar_atom1._x, 3),
#               print round(polar_atom1._y, 3),
#               print round(polar_atom1._z, 3)
                for polar_atom2 in polar_neighbors:
                    if polar_atom1 is polar_atom2:
                        # Same atom
                        continue
                    midpoints.append(min_pol_rad_pointA(
                        polar_atom1,
                        polar_atom2,
                        ( polar_atom1.char_pol_radius
                            - (polar_atom1.char_pol_radius_max
                                + polar_atom1.char_pol_radius ))))
            if len(midpoints) < 2:
                continue
            center = Point(0.0,0.0,0.0)
            for midpoint in midpoints:
                center = center + midpoint
            center = center / len(midpoints)
#           print "@balllist {add_void_waters1} colour=yellow radius=0.1"
#           print "{ }",
#           print round(center._x, 3),
#           print round(center._y, 3),
#           print round(center._z, 3)

            # Check for clashes with neighbors
            insert_water = True
            for neighbor in probe._far_neighbors:
                if neighbor._sub_atom_type == 4:
                    continue
                cutoff = neighbor._close_radius
                if neighbor._atom_polarity == 0:
                    cutoff = ( min_water_radius
                        + neighbor.char_pol_radius
                        - ( neighbor.char_pol_radius_max
                            - neighbor.char_pol_radius ) )
                if center.clashes(neighbor, cutoff):
                    insert_water = False
                    break
            if insert_water:
                self._hydration.atom_input(
                    -1,                 # hydration_layer
                    res_start_id,               # resOrgId
                    None,               # atomOriginalId
                    ' O  ',
                    'O',
                    center._x,
                    center._y,
                    center._z,
                    3,                  # subAtomType
                    1)                  # moleculeInstance
                update_neighbors(
                    self._hydration[-1],
                    dynamic_neighbor_list)
                dynamic_neighbor_list.append(self._hydration[-1])
                self._hydration[-1].set_water_type(0.1)
                res_start_id += 1
#               print "@balllist {add_void_waters1} colour=yellow radius=0.1"
#               print "{ }",
#               print round(center._x, 3),
#               print round(center._y, 3),
#               print round(center._z, 3)

        return res_start_id

    def add_void_waters2(self, res_start_id, probe_list, all_atoms):
        """ function add_void_waters2

        res_start_id: int
        probe_list: list
        all_atoms: list

        This is a method for checking if waters can be placed in polar
        cavities, by using the minimum polar radius of water and the
        neighboring polar atoms. The apolar atoms use the closeRadius.

        The method loops through the VoidProbes in the  probe_list. For each
        probe it checks that it has got at least one polar close neighbor. It
        then checks for clashes with the neighboring atoms (far neighbors),
        using the cutoff described above. If there are no clashes it places
        the water updates the neighbor lists and sets the accessible surface
        of the water.

        """
        dynamic_neighbor_list = all_atoms[:]
        min_water_radius = R.water_radius - ( R.water_radius_max - R.water_radius )
        for probe in probe_list:
            # This flag is used to determine wheather or not the probe has any
            # polar neighbors
            num_polar_neighbors = 0
            for close_neighbor in probe._close_neighbors:
                if close_neighbor._sub_atom_type != 4:
                    if close_neighbor._atom_polarity == 0:
                        num_polar_neighbors = num_polar_neighbors + 1
            if num_polar_neighbors < 1:
                continue
                
            # This flag is used to determine wheather or not to place a water
            # at the probe center
            place_water = True
            # Loop through the probe neighbors
            for neighbor in probe._far_neighbors:
                if neighbor._sub_atom_type == 4:
                    continue
                # Defualt cutoff to apolar neighbors
                cutoff = neighbor._close_radius
                if neighbor._atom_polarity == 0:
                    # Change cutoff for polar neighbors
                    cutoff = ( min_water_radius
                        + neighbor.char_pol_radius
                        - ( neighbor.char_pol_radius_max
                            - neighbor.char_pol_radius ) )
                if probe.clashes(neighbor, cutoff):
                    place_water = False
                    break
            if place_water:
                self._hydration.atom_input(
                    -1,                 # hydration_layer
                    res_start_id,               # resOrignalId
                    None,               # atomOriginalId
                    ' O  ',
                    'O',
                    probe._x,
                    probe._y,
                    probe._z,
                    3,                  # subAtomType
                    1)                  # moleculeInstance
                update_neighbors(
                    self._hydration[-1],
                    dynamic_neighbor_list)
                dynamic_neighbor_list.append(self._hydration[-1])
                self._hydration[-1].set_water_type(0.1)
                res_start_id += 1
    
        return res_start_id

    def set_contacts(self, atom_list, target_list):
        """function set_contacts

        atom_list: list
        target_list: list
        
        This method saves contacts into the ContactList container.
        
        It loops through all the atoms in the MacroMolecule container
        and checks for apolar and polar contacts to atoms in the farNeigbors
        list. Contacts within a residue are ignored, contacts between
        peptide bond atoms are ignored.

        It also loops through all the atoms in the Ligand container and
        checks for apolar and polar contacts to atoms in the farNeighbors
        list. Contacts to LigandAtoms are ignored.
        
        """
        interactions = self.interaction
        if interactions == None:
            raise AssemblyError, "set_contacts, interaction not set"
#       for atom in atom_list:
#           for target in target_list:
#               contact = atom.in_contact(target)
#               if contact != -1:
#                   # Add to contactList
#                   interactions.contact_input(atom, target, contact)
        for atom in atom_list:
            for target in atom._close_neighbors:
                if atom not in target_list:
                    continue
                contact = atom.in_contact(target)
                if contact != -1:
                    # Add to contactList
                    interactions.contact_input(atom, target, contact)
    
    def set_binding_site_using_cutoff(self, unit_points, cutoff):
        """
        set_binding_site_using_cutoff
        
        """
        cutoff2 = cutoff*cutoff
    
        protein_list = self.macro_molecule
        ligand_list = self.ligand
        for protein_atom in protein_list:
            protein_atom.include = 0
        for ligand_atom in ligand_list:
            for protein_atom in protein_list:
                if protein_atom._include == 1:
                    continue
                d2 = ligand_atom.get_distance2(protein_atom)
                if d2 < cutoff2:
                    protein_atom.include = 1

    def set_binding_site(self, unit_points):
        """function set_binding_site

        All protein atoms in direct contact with the ligand are classified as
        being part of the binding site.

        Theoretical waters are placed on all solvent accessible surface points
        of ligand atoms.  Any protein atoms in direct contact with the
        theoretical waters are also part of the binding site.

        Probes (allowed to move and expand) are added to ligand probe
        accessible surfaces. Any protein atoms in direct contact with a probe
        is also part of the binding site.
        
        """
        # Create the protein, ligand and protein-ligand complex atom lists
        protein_list = self.macro_molecule
        ligand_list = self.ligand
        complex_list = protein_list + ligand_list
        
        # The defualt Atom.include attribute is 1
        # Set include to zero for all the protein atoms
        for protein_atom in protein_list:
            protein_atom.include = 0
        
        # Set include to 1 for all the protein atoms in contact with
        # the ligand
        for ligand_atom in ligand_list:
            for protein_atom in protein_list:
                if protein_atom._include == 1:
                    continue
                if ligand_atom.in_contact(protein_atom) != -1:
                    protein_atom.include = 1
        
        # Set the neighbors of the protien and ligand atoms as they would
        # appear in the complex
        set_neighbors(complex_list)
        for atom in complex_list:
            atom.set_surface_points(unit_points, atom._close_radius, 0, 0)
            atom.set_accessibility(unit_points)
        
        # Add waters to the ligand (in the complex)
        tmp_waters = []
        for ligand_atom in ligand_list:
            for surface_point in ligand_atom._surface._surface_point_list:
                if ( surface_point._accessibility == 0
                    or surface_point._accessibility == 2 ):
                    tmp_waters.append(WaterProbe(
                        1,
                        ' O  ',
                        'H2O',
                        surface_point._x,
                        surface_point._y,
                        surface_point._z,
                        3,
                        0,
                        1,
                        False))
                    # below is a hack for the get_kin_waters function to work
                    # on the tmp_waters list
                    tmp_waters[-1]._water_type = 0
                    tmp_waters[-1]._hydration_layer = 1
        
        # Set include to 1 for all the protein atoms in contact with
        # water atoms
        for water_atom in tmp_waters:
            for protein_atom in protein_list:
                if protein_atom._include == 1:
                    continue
                if protein_atom.in_contact(water_atom) != -1:
                    protein_atom.include = 1
        
        # Fill voids with probes
        self.probe_fill(unit_points, ligand_list)
        
        # Set include to 1 for all the protein atoms in contact with
        # probes
        probe_list = self.void
        for probe_atom in probe_list:
            for protein_atom in protein_list:
                if protein_atom._include == 1:
                    continue
                if protein_atom.in_contact(probe_atom) != -1:
                    protein_atom.include = 1
        
        if __debug__:
            # Print the waters
            bs_waters_file_name = "binding_site_waters.kin"
            bs_waters_file = open(bs_waters_file_name, mode="w")
            bs_waters_text = self.get_kin_waters(tmp_waters)
            bs_waters_file.write(bs_waters_text)
            bs_waters_file.close()

            # print the cavity defining pseudo-atoms
            bs_probes_file_name = "binding_site_probes.kin"
            bs_probes_file = open(bs_probes_file_name, mode="w")
            bs_probes_text = self.get_kin_probes(probe_list)
            bs_probes_file.write(bs_probes_text)
            bs_probes_file.close()

        # Clear the void and hydration
        self.void.clear()
        self.hydration.clear()
        
        # Clear all neighbor references
        for atom in (self.ligand + self.macro_molecule):
            atom.clear_neighbors()
            if atom._surface != None:
                atom.clear_surface()

    def set_interface(self, unit_points):
        """function set_interface
        
        unit_points: UnitPoints

        This method excludes all non-interface protein atoms from the rest of
        the analysis by setting their "include" attribute to 0.

        The interface is defined by adding one hydration layer to one protein.
        Any protein atom from the other subunit in direct contact with the
        first subunit, or with a water molecule, or with a probe is defined as
        being part of the interface. The same process is then repeated
        inversing the protein subunits.
        
        """
        # Create the subunit atom lists
        all_protein_atoms = self.macro_molecule
        subunit1 = []
        subunit2 = []
        for atom in all_protein_atoms:
            if atom._molecule_instance == 1:
                subunit1.append(atom)
            elif atom._molecule_instance == 2:
                subunit2.append(atom)

        # First subunit1
        # Set the neighbors of the protein atoms as they would appear
        # in the complex
        set_neighbors(all_protein_atoms)

        for atom in all_protein_atoms:
            atom.set_surface_points(unit_points, atom._close_radius, 0, 0)
            atom.set_accessibility(unit_points)
        
        # Add waters
        tmp_waters = []
        for atom in subunit1:
            for surface_point in atom._surface._surface_point_list:
                if ( surface_point._accessibility == 0
                or surface_point._accessibility == 2):
                    tmp_waters.append(WaterProbe(
                        1,
                        ' O  ',
                        'H2O',
                        surface_point._x,
                        surface_point._y,
                        surface_point._z,
                        3,
                        0,
                        1,
                        False))

        # Fill voids with probes
        self.probe_fill(unit_points, subunit1)

        # The default Atom.include attribute is 1
        # Set include to zero for the subunit protein atoms
        for protein_atom in subunit2:
            protein_atom.include = 0
        
        # Set include to 1 for all the subunit atoms in contact with
        # the other subuit
        for atom_to_touch in subunit1:
            for atom_to_include in subunit2:
                if atom_to_include._include == 1:
                    continue
                if atom_to_touch.in_contact(atom_to_include) != -1:
                    atom_to_include.include = 1

        for water_atom in tmp_waters:
            for atom_to_include in subunit2:
                if atom_to_include._include == 1:
                    continue
                if atom_to_include.in_contact(water_atom) != -1:
                    atom_to_include.include = 1
        
        probe_list = self.void
        for probe_atom in probe_list:
            for atom_to_include in subunit2:
                if atom_to_include._include == 1:
                    continue
                if atom_to_include.in_contact(probe_atom) != -1:
                    atom_to_include.include = 1
        
        # Clear the tmp_waters
        del tmp_waters[:]
        del tmp_waters
        
        # Clear the Assembly.void
        for atom in self.void:
            atom.clear_neighbors()
        self.void.clear()
        
        # Second subunit2
        # Add waters
        tmp_waters = []
        for atom in subunit2:
            for surface_point in atom._surface._surface_point_list:
                if ( surface_point._accessibility == 0
                or surface_point._accessibility == 2):
                    tmp_waters.append(WaterProbe(
                        1,
                        ' O  ',
                        'H2O',
                        surface_point._x,
                        surface_point._y,
                        surface_point._z,
                        3,
                        0,
                        1,
                        False))

        # Fill voids with probes
        self.probe_fill(unit_points, subunit2)

        # The default Atom.include attribute is 1
        # Set include to zero for the subunit protein atoms
        for protein_atom in subunit1:
            protein_atom.include = 0
        
        # Set include to 1 for all the subunit atoms in contact with
        # the other subuit
        for atom_to_touch in subunit2:
            for atom_to_include in subunit1:
                if atom_to_include._include == 1:
                    continue
                if atom_to_touch.in_contact(atom_to_include) != -1:
                    atom_to_include.include = 1

        for water_atom in tmp_waters:
            for atom_to_include in subunit1:
                if atom_to_include._include == 1:
                    continue
                if atom_to_include.in_contact(water_atom) != -1:
                    atom_to_include.include = 1
        
        probe_list = self.void
        for probe_atom in probe_list:
            for atom_to_include in subunit1:
                if atom_to_include._include == 1:
                    continue
                if atom_to_include.in_contact(probe_atom) != -1:
                    atom_to_include.include = 1

        # Clear the tmp_waters
        del tmp_waters[:]
        del tmp_waters
        
        # Clear the Assembly.void
        self.void.clear()
        self.hydration.clear()
        
        # Clear all neighbor references
        for atom in self.macro_molecule:
            atom.clear_neighbors()
            if atom._surface != None:
                atom.clear_surface()
    
    def get_kin_accessibility(self, atom_list):
        """function get_kin_accessibility
        
        atom_list: list

        This method returns a text string. The text string can be
        written to a .kin file and visulised using mage.

        The kin text string contains;
        1) A description of the file  (@text...)
        2) Vectors going from the atom to the accessible point
           colour coded by the accessibility type
        3) Dot points at the accessible points colour coded by the
           accessibility type
        
        returns string
        """
        kin_list = []
        info_text = """@text
                    This file contains the atom surface areas.
                    These can be probe inaccessible, probe accessible,
                    polar solvent accessibe and apolar solvent accessible.
                    """
        kin_list.append(re.sub(r'\n\s+', '\n', info_text))
        kin_list.append("@kinemage")
        kin_list.append("master={SAS}")
        kin_list.append("master={PSAS}")
        kin_list.append("master={SIAS}")
        kin_list.append("@vectorlist {PolSAS 0} color=red master={SAS}")
        atom_id = 0
        for atom in atom_list:
            atom_id = atom_id + 1
            point_id = 0
            if atom._include != 1:
                continue
            surface = atom._surface
            for surface_point in surface._surface_point_list:
                point_id = point_id + 1
                if surface_point._accessibility == 0:
                    tmp_kin = "{} P %.4f %.4f %.4f {%i_%i} %.4f %.4f %.4f" % (
                                atom._x,
                                atom._y,
                                atom._z,
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@vectorlist {ApolSAS 2} color=green master={SAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == 2:
                    tmp_kin = "{} P %.4f %.4f %.4f {%i_%i} %.4f %.4f %.4f" % (
                                atom._x,
                                atom._y,
                                atom._z,
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@vectorlist {PSAS -1} color=magenta off master={PSAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == -1:
                    tmp_kin = "{} P %.4f %.4f %.4f {%i_%i} %.4f %.4f %.4f" % (
                                atom._x,
                                atom._y,
                                atom._z,
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@vectorlist {PSAS -3} color=purple off master={PSAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == -3:
                    tmp_kin = "{} P %.4f %.4f %.4f {%i_%i} %.4f %.4f %.4f" % (
                                atom._x,
                                atom._y,
                                atom._z,
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@vectorlist {SIAS 1} color=cyan off master={SIAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == 1:
                    tmp_kin = "{} P %.4f %.4f %.4f {%i_%i} %.4f %.4f %.4f" % (
                                atom._x,
                                atom._y,
                                atom._z,
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@vectorlist {SIAS -2} color=sky off master={SIAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == -2:
                    tmp_kin = "{} P %.4f %.4f %.4f {%i_%i} %.4f %.4f %.4f" % (
                                atom._x,
                                atom._y,
                                atom._z,
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@dotlist {PolSAS 0} color=red master={SAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == 0:
                    tmp_kin = "{%i_%i} %.4f %.4f %.4f" % (
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@dotlist {ApolSAS 2} color=green master={SAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == 2:
                    tmp_kin = "{%i_%i} %.4f %.4f %.4f" % (
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@dotlist {PSAS -1} color=magenta off master={PSAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == -1:
                    tmp_kin = "{%i_%i} %.4f %.4f %.4f" % (
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@dotlist {PSAS -3} color=purple off master={PSAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == -3:
                    tmp_kin = "{%i_%i} %.4f %.4f %.4f" % (
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@dotlist {PSIAS 1} color=cyan off master={SIAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == 1:
                    tmp_kin = "{%i_%i} %.4f %.4f %.4f" % (
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        kin_list.append("@dotlist {PSIAS -2} color=sky off master={SIAS}")
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include != 1:
                continue
            surface = atom._surface
            for point_id, surface_point in enumerate(surface._surface_point_list):
                point_id = point_id + 1
                if surface_point._accessibility == -2:
                    tmp_kin = "{%i_%i} %.4f %.4f %.4f" % (
                                atom_id,
                                point_id,
                                surface_point._x,
                                surface_point._y,
                                surface_point._z)
                    kin_list.append(tmp_kin)
        return "\n".join(kin_list)
    
    def get_naccess_accessibility(self, unit_points, atom_list):
        """
        Return string  representing naccess output.

        """
        residue_data = {}
        num_surface_points = float(len(unit_points.point_list))
        for atom in atom_list:
            if atom._include != 1:
                continue
            surface = atom._surface
            # Deal with ligands
            try:
                res_id = atom._res_original_id
            except AttributeError:
                res_id = "LIG"
            atom_name = atom._atom_org_name
            atom_id = atom._atom_org_id
            tot_sp = 0
            residue_data.setdefault(res_id, {}).setdefault(
                                    atom_id, {})
            area = 0.0
            for surface_point in surface._surface_point_list:
                acc = surface_point._accessibility
                if acc != 0 and acc != 2:
                    continue
                tot_sp += 1
                r = surface_point._radius
                partial_area = 4*pi*r*r / num_surface_points
                area += partial_area
            frac_solvent_exposed = tot_sp/num_surface_points
            residue_data[res_id][atom_id] = (atom, area)
        res_keys = residue_data.keys()
        res_keys.sort()
        pdb_list = []
        rsa_list = []
        rsa_list.append('REM SOLVENT ACCESSIBILITY CALCULATED BY PRO_ACT2')
        rsa_list.append('REM RESIDUE SOLVENT ACCESSIBILITY')
        rsa_list.append('REM RES _  NUM       TOT    APOLAR     POLAR')
        tot_pol = 0.0
        tot_apol = 0.0
        chain_pol = {}
        chain_apol = {}
        for res_key in res_keys:
            res_pol = 0.0
            res_apol = 0.0
            atoms = residue_data[res_key]
            atom_keys = atoms.keys()
            atom_keys.sort()
            res_type = None
            chain = None
            res_original_id = None
            rsa_field = None
            for atom_key in atom_keys:
                atom = residue_data[res_key][atom_key][0]
                atom_id = atom._atom_org_id
                area = residue_data[res_key][atom_key][1]
                # Deal with ligands
                try:
                    field = 'ATOM'
                    res_type = atom._res_type
                    chain = atom._chain
                    res_original_id = atom._res_original_id
                except AttributeError:
                    field = 'HETATM'
                    res_type = "LIG"
                    chain = "Z"
                    res_original_id = 9999
                if atom._atom_polarity == 0:
                    res_pol += area
                    chain_pol[chain] = chain_pol.setdefault(chain, 0.0) + area
                    chain_apol.setdefault(chain, 0.0)
                elif atom._atom_polarity == 2:
                    res_apol += area
                    chain_apol[chain] = chain_apol.setdefault(chain, 0.0) + area
                    chain_pol.setdefault(chain, 0.0)
                pdb_line = '%-6s%5i %4s %3s %1s%4i    %8.3f%8.3f%8.3f%6.2f%6.2f           %1s' % (
                        field,
                        atom_id,
                        atom._atom_org_name,
                        res_type,
                        chain,
                        res_original_id,
                        atom._x,
                        atom._y,
                        atom._z,
                        1.0,
                        area,
                        atom._atom_type)
                pdb_list.append(pdb_line)
            res_tot = res_pol + res_apol
            tot_pol += res_pol
            tot_apol += res_apol

            rsa_line = '%3s %3s %1s %4i %9.2f %9.2f %9.2f' % (
                        'RES',
                        res_type,
                        chain,
                        res_original_id,
                        res_tot,
                        res_apol,
                        res_pol)
            rsa_list.append(rsa_line)
        rsa_list.append('END')
        rsa_list.append('REM CHAIN SOLVENT ACCESSIBILITY')
        rsa_list.append('REM                  TOT    APOLAR     POLAR')
        chain_keys = chain_apol.keys()
        chain_keys.sort()
        for key in chain_keys:
            chain_tot = chain_apol[key] + chain_pol[key]
            rsa_line = 'CHAIN   %1s      %9.2f %9.2f %9.2f' % (
                        key,
                        chain_tot,
                        chain_apol[key],
                        chain_pol[key])
            rsa_list.append(rsa_line)
        rsa_list.append('END')
        rsa_list.append('REM TOTAL SOLVENT ACCESSIBILITY')
        rsa_list.append('REM                  TOT    APOLAR     POLAR')
        tot = tot_pol + tot_apol
        rsa_line = 'TOTAL          %9.2f %9.2f %9.2f' % (
                    tot,
                    tot_apol,
                    tot_pol)
        rsa_list.append(rsa_line)
        rsa_list.append('END')
        pdb_text = "\n".join(pdb_list)
        rsa_text = "\n".join(rsa_list)
        return (pdb_text, rsa_text)

    def get_dms_surface(self, unit_points, atom_list):
        """
        Return string representing ms output.

        """
        lines = []
        residue_data = {}
        num_surface_points = float(len(unit_points.point_list))
        for atom in atom_list:
            if atom._include != 1:
                continue
            surface = atom._surface
            # Deal with ligands
            try:
                res_id = atom._res_original_id
            except AttributeError:
                res_id = "LIG"
            atom_name = atom._atom_org_name
            atom_id = atom._atom_org_id
            residue_data.setdefault(res_id, {}).setdefault(
                                    atom_id, {})
            residue_data[res_id][atom_id] = atom
        res_keys = residue_data.keys()
        res_keys.sort()
        for res_key in res_keys:
            atoms = residue_data[res_key]
            atom_keys = atoms.keys()
            atom_keys.sort()
            for atom_key in atom_keys:
                atom = residue_data[res_key][atom_key]
                surface = atom._surface
                atom_name = atom._atom_org_name.strip()
                line = '%s%4i%c %4s%8.3f %8.3f %8.3f A' % (atom._res_type,
                                                             res_key,
                                                             atom._chain,
                                                             atom_name,
                                                             atom._x,
                                                             atom._y,
                                                             atom._z)
                lines.append(line)
                for surface_point in surface._surface_point_list:
                    acc = surface_point._accessibility
                    if acc != 0 and acc != 2:
                        continue
                    r = surface_point._radius
                    partial_area = 4*pi*r*r / num_surface_points
                    line = '%s%4i%c %4s%8.3f %8.3f %8.3f SC0%7.3f%7.3f%7.3f%7.3f' % (
                                                             atom._res_type,
                                                             res_key,
                                                             atom._chain,
                                                             atom_name,
                                                             surface_point._x,
                                                             surface_point._y,
                                                             surface_point._z,
                                                             partial_area,
                                                             surface_point._unit_surface_point.x,
                                                             surface_point._unit_surface_point.y,
                                                             surface_point._unit_surface_point.z)
                    lines.append(line)
        return '\n'.join(lines)

    def get_kin_waters(self, water_probe_list):
        """function get_kin_waters
        
        water_probe_list: list

        This method returns a text sting. The text string can be
        written to a .kin file and visualised using mage.

        The kin text string contains:
        1) A description of the file (@text...)
        2) Balls representing the WaterProbes colour coded by WaterType
        
        returns string
        """
        kin_list = []
        info_text = """@text
                    This file contains the water atoms. Waters can be:
                    Buried  - 0 or -1
                    Cleft   - 2, 3, ..., n
                    Surface - 1
                    Further - -1, -2, ..., -n
                    """
        kin_list.append(re.sub(r'\n\s+', '\n', info_text))
        kin_list.append("@kinemage")
        kin_list.append("@spherelist {Buried} color=gold radius=%.2f" % R.water_radius)
        for water_id, water in enumerate(water_probe_list):
            water_id = water_id + 1
            if water._water_type == 0:
                tmp_kin = "{%i hydLay=%i watType=%i experimental=%s} %.4f %.4f %.4f" % (
                            water_id,
                            water.hydration_layer,
                            water._water_type,
                            str(water.experimental),
                            water._x,
                            water._y,
                            water._z)
                kin_list.append(tmp_kin)
        kin_list.append("@spherelist {Surface} color=pinktint radius=%.2f" % R.water_radius)
        for water_id, water in enumerate(water_probe_list):
            water_id = water_id + 1
            if water._water_type == 1:
                tmp_kin = "{%i hydLay=%i watType=%i experimental=%s} %.4f %.4f %.4f" % (
                            water_id,
                            water.hydration_layer,
                            water._water_type,
                            str(water.experimental),
                            water._x,
                            water._y,
                            water._z)
                kin_list.append(tmp_kin)
        kin_list.append("@spherelist {Cleft} color=peachtint radius=%.2f" % R.water_radius)
        for water_id, water in enumerate(water_probe_list):
            water_id = water_id + 1
            if water._water_type > 1:
                tmp_kin = "{%i hydLay=%i watType=%i experimental=%s} %.4f %.4f %.4f" % (
                            water_id,
                            water.hydration_layer,
                            water._water_type,
                            str(water.experimental),
                            water._x,
                            water._y,
                            water._z)
                kin_list.append(tmp_kin)
        kin_list.append("@spherelist {Further} color=lilactint radius=%.2f" % R.water_radius)
        for water_id, water in enumerate(water_probe_list):
            water_id = water_id + 1
            if water._water_type < 0:
                tmp_kin = "{%i hydLay=%i watType=%i experimental=%s} %.4f %.4f %.4f" % (
                            water_id,
                            water.hydration_layer,
                            water._water_type,
                            str(water.experimental),
                            water._x,
                            water._y,
                            water._z)
                kin_list.append(tmp_kin)
        return "\n".join(kin_list)

    def get_pdb_waters(self, water_probe_list):
        """
        Return string of waters in PDB file formtat.

        water_probe_list: list

        This method returns a text string containing all the waters in PDB
        file format. Experimental waters get the B-factor value set to 1.
        Theoretical waters get the occupancy value set to 0. Buried waters
        have their chain set to 0. Surface waters have their chain value set
        to 1. Cleft waters have their value set to > 1. Further waters have
        their chain value set to F.

#       This method also sets the water atom_ids, which are used by the
#       get_residue_residue_contects method.

        returns string
        """
        pdb_list = []
        data = []
        for water in water_probe_list:
            if water.water_type < 0:
                water_type = 'F'
            else:
                water_type = '%i' % water.water_type
            # As False comes before True in a sort we need to give the
            # 'experimetal' dummy variable some values where the experimental
            # waters gets sorted before the theoretical waters
            if water.experimental:
                experimental = 0
            else:
                experimental = 1
            datum = (water_type, experimental, water)
            data.append(datum)
#       data.sort()
        water_id = 0
        for datum in data:
            water_id += 1
#           water._atom_org_id = water_id  # Set the id of the water molecule
            water_type = datum[0]
            experimental = datum[1]
            water = datum[2]
            # Correct the value used for sorting so that experimental waters
            # have an occupancy of 1 and theoretical waters have an occupancy
            # of 0
            if experimental == 0:
                experimental = 1
            elif experimental == 1:
                experimental = 0
            line = 'HETATM%5i %4s %3s %s%4i    %8.3f%8.3f%8.3f %5.2f %5.2f' % (
                        water_id,
                        ' O  ',
                        'HOH',
                        water_type,
                        water._res_original_id,
                        water._x,
                        water._y,
                        water._z,
                        experimental,
                        0.0)
            pdb_list.append(line)
            
        return '\n'.join(pdb_list)
            
    def get_kin_probes(self, void_probe_list):
        """function get_kin_probes
        
        void_probe_list: list

        This method returns a text sting. The text string can be
        written to a .kin file and visualised using mage.

        The kin text string contains:
        1) A description of the file (@text...)
        2) Balls representing the VoidProbes
        
        returns string
        """
        kin_list = []
        info_text = """@text
                    This file contains the cavity-defining pseudo-atoms.
                    """
        kin_list.append(re.sub(r'\n\s+', '\n', info_text))
        kin_list.append("@kinemage")
        kin_list.append("@spherelist {Probes} color=purple")
        for probe_id, probe in enumerate(void_probe_list):
            probe_id = probe_id + 1
            tmp_kin = "{ %i r=%.4f } %.4f %.4f %.4f r=%.4f" % (
                        probe_id,
                        probe._probe_radius,
                        probe._x,
                        probe._y,
                        probe._z,
                        probe._probe_radius)
            kin_list.append(tmp_kin)
        return "\n".join(kin_list)

    def get_pdb_probes(self, void_probe_list):
        """
        Return string of cavity probes in PDB file format.
        
        void_probe_list: list

        This method returns a text string containing all the cavity probes in
        PDB file format. The B-factor value represents the radius of a
        particular cavity probe. The atom type is set to carbon ' C  ', and
        the HETATM name to CCC. The chain is set to C.
        
        returns string
        """
        pdb_list = []
        data = []
        chain = 'C'
        for probe_id, probe in enumerate(void_probe_list):
            probe_id = probe_id + 1
            radius = probe._probe_radius
            line = 'HETATM%5i %4s %3s %s%4i    %8.3f%8.3f%8.3f %5.2f %5.2f' % (
                        probe_id,
                        ' C  ',
                        'CCC',
                        chain,
                        probe_id,
                        probe._x,
                        probe._y,
                        probe._z,
                        0.0,
                        radius)
            pdb_list.append(line)
            
        return '\n'.join(pdb_list)

    def get_kin_include(self, atom_list):
        """function get_kin_include

        atom_list: list
        
        This method returns a text sting. The text string can be
        written to a .kin file and visualised using mage.

        The kin text string contains:
        1) A description of the file (@text...)
        2) Balls representing the atoms that have the atom.include attribute
           set to 1
        
        returns string
        """
        kin_list = []
        info_text = """@text
                    This file contains the atoms that have had their atom.include attribute
                    switched to 1.
                    """
        kin_list.append(re.sub(r'\n\s+', '\n', info_text))
        kin_list.append("@kinemage")
        carbon_list = []
        oxygen_list = []
        nitrogen_list = []
        sulphur_list = []
        other_list = []
        carbon_radius = 0.0
        oxygen_radius = 0.0
        nitrogen_radius = 0.0
        sulphur_radius = 0.0
        other_radius = 0.0
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            if atom._include == 1:
                radius = atom._char_apol_radius
                if atom._atom_polarity == 0:
                    radius = atom.char_pol_radius
                tmp_kin = "{ atmName=%s atmType=%s atomId=%i chain=%s} %.4f %.4f %.4f" % (
                            atom._atom_org_name,
                            atom._atom_type,
                            atom._res_original_id,
                            atom.chain,
                            atom._x,
                            atom._y,
                            atom._z)
                if atom._atom_type == 'C':
                    carbon_list.append(tmp_kin)
                    carbon_radius = radius
                elif atom._atom_type == 'O':
                    oxygen_list.append(tmp_kin)
                    oxygen_radius = radius
                elif atom._atom_type == 'N':
                    nitrogen_list.append(tmp_kin)
                    nitrogen_radius = radius
                elif atom._atom_type == 'S':
                    sulphur_list.append(tmp_kin)
                    sulphur_radius = radius
                else:
                    other_list += tmp_kin
                    other_radius = radius
        if carbon_list != "":
            h = "@spherelist {Carbon} color=gray radius=%.2f master={Binding site}" % carbon_radius
            kin_list.append(h)
            kin_list.extend(carbon_list)
        if oxygen_list != "":
            h = "@spherelist {Oxygen} color=red radius=%.2f master={Binding site}" % oxygen_radius
            kin_list.append(oxygen_header)
            kin_list.extend(oxygen_list)
        if nitrogen_list != "":
            h = "@spherelist {Nitrogen} color=blue radius=%.2f master={Binding site}" % nitrogen_radius
            kin_list.append(h)
            kin_list.extend(nitrogen_list)
        if sulphur_list != "":
            h = "@spherelist {Sulphur} color=yellow radius=%.2f master={Binding site}" % sulphur_radius
            kin_list.append(h)
            kin_list.extend(sulphur_list)
        if other_list != "":
            h = "@spherelist {Other} color=magenta radius=%.2f master={Binding site}" % other_radius
            kin_list.append(h)
            kin_list.extend(other_list)

        return "\n".join(kin_list)
        
    def get_kin_ligand(self, atom_list):
        """function get_kin_ligand
        
        This method returns a kin string.

        returns string
        """
        kin_list = []
        info_text = """@text
                    This file contains the ligand.
                    """
        kin_list.append(re.sub(r'\n\s+', '\n', info_text))
        kin_list.append("@kinemage")
        carbon_list = []
        oxygen_list = []
        nitrogen_list = []
        sulphur_list = []
        halogen_list = []
        phosphor_list = []
        other_list = []
        carbon_radius = 0.0
        oxygen_radius = 0.0
        nitrogen_radius = 0.0
        sulphur_radius = 0.0
        halogen_radius = 0.0
        phosphor_radius = 0.0
        other_radius = 0.0
        for atom_id, atom in enumerate(atom_list):
            atom_id = atom_id + 1
            radius = atom._char_apol_radius
            if atom._atom_polarity == 0:
                radius = atom.char_pol_radius
            tmp_kin = "{ %i %s r=%.2f } %.4f %.4f %.4f" % (
                        atom_id,
                        atom_type, 
                        radius,
                        atom._x,
                        atom._y,
                        atom._z)
            if atom._atom_type == 'C':
                carbon_list.append(tmp_list)
                carbon_radius = radius
            elif atom._atom_type == 'O':
                oxygen_list.append(tmp_list)
                oxygen_radius = radius
            elif atom._atom_type == 'N':
                nitrogen_list.append(tmp_list)
                nitrogen_radius = radius
            elif atom._atom_type == 'S':
                sulphur_list.append(tmp_list)
                sulphur_radius = radius
            elif ( atom._atom_type == 'F'
                or atom._atom_type == 'Cl'
                or atom._atom_type == 'Br'
                or atom._atom_type == 'I' ):
                halogen_list.append(tmp_list)
                halogen_radius = radius
            elif atom._atom_type == 'P':
                phosphor_list.append(tmp_list)
                phosphor_radius = radius
            else:
                other_list.append(tmp_list)
                other_radius = radius
        if len(carbon_list) > 0:
            h = "@spherelist {Carbon} color=peach radius=%.2f master={Ligand}" % carbon_radius
            kin_list.append(h)
            kin_list.extend(carbon_list)
        if len(oxygen_list) > 0:
            h = "@spherelist {Oxygen} color=pink radius=%.2f master={Ligand}" % oxygen_radius
            kin_list.append(h)
            kin_list.extend(oxygen_list)
        if len(nitrogen_list) > 0:
            h = "@spherelist {Nitrogen} color=sky radius=%.2f master={Ligand}" % nitrogen_radius
            kin_list.append(nitrogen_header)
            kin_list.extend(nitrogen_list)
        if len(sulphur_list) > 0:
            h = "@spherelist {Sulphur} color=yellowtint radius=%.2f master={Ligand}" % sulphur_radius
            kin_list.append(sulphur_header)
            kin_list.extend(sulphur_list)
        if len(halogen_list) > 0:
            h = "@spherelist {Halogen} color=sea radius=%.2f master={Ligand}" % halogen_radius
            kin_list.append(halogen_header)
            kin_list.extend(halogen_list)
        if len(phosphor_list) > 0:
            h = "@spherelist {Phosphor} color=lilactint radius=%.2f master={Ligand}" % phosphor_radius
            kin_list.append(phosphor_header)
            kin_list.extend(phosphor_list)
        if len(other_list) > 0:
            h = "@spherelist {Other} color=magenta radius=%.2f master={Ligand}" % other_radius
            kin_list.append(other_header)
            kin_list.extend(other_list)
        return "\n".join(kin_list)

    def get_kin_contacts(self):
        """function get_kin_contacts

        This method returns a text string. The text string can be written to a
        .kin file and visualised using mage.

        Loops through all the contacts and prints them.

        return string
        """
        kin_list = []
        info_text = """@text
                    This file contains vectors representing:
                    1) Polar contacts
                    2) Apolar contacts
                    3) Polar water contacts
                    4) Apolar water contacts
                    5) Polar void contacts
                    6) Apolar void contacts
                    as well as dots representing the midpoint of contacts.
                    """
        kin_list.append(re.sub(r'\n\s+', '\n', info_text))
        kin_list.append("@kinemage")
        contacts = self.interaction

        polar_contacts = []
        apolar_contacts = []
        polar_water_contacts = []
        apolar_water_contacts = []
        polar_void_contacts = []
        apolar_void_contacts = []
        polar_contact_dots = []
        apolar_contact_dots = []
        polar_water_contact_dots = []
        apolar_water_contact_dots = []
        polar_void_contact_dots = []
        apolar_void_contact_dots = []

        polar_contacts.append("@vectorlist {Polar Contacts} color=red width=3")
        apolar_contacts.append("@vectorlist {Apolar Contacts} color=green width=3")
        polar_water_contacts.append("@vectorlist {Polar Water Contacts} color=red off")
        apolar_water_contacts.append("@vectorlist {Apolar Water Contacts} color=green off")
        polar_void_contacts.append("@vectorlist {Polar Void Contacts} color=red off")
        apolar_void_contacts.append("@vectorlist {Apolar Void Contacts} color=green off")
        polar_contact_dots.append("@dotlist {Polar Contacts} color=red width=3")
        apolar_contact_dots.append("@dotlist {Apolar Contacts} color=green width=3")
        polar_water_contact_dots.append("@dotlist {Polar Water Contacts} color=red off")
        apolar_water_contact_dots.append("@dotlist {Apolar Water Contacts} color=green off")
        polar_void_contact_dots.append("@dotlist {Polar Void Contacts} color=red off")
        apolar_void_contact_dots.append("@dotlist {Apolar Void Contacts} color=green off")

        for contact in contacts:
            atom1 = contact._atom1
            atom2 = contact._atom2
            polarity = contact._contact_polarity
            type = 0
            tmp_string = ""
            if atom1._sub_atom_type == 1:
                # atom1 is a protein atom
                tmp_string = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f" % (
                                atom1._atom_org_name,
                                atom1._atom_type,
                                atom1._res_original_id,
                                atom1._res_type,
                                atom1._chain,
                                atom1._x,
                                atom1._y,
                                atom1._z)
                if atom2._sub_atom_type == 1:
                    # atom2 is a protein atom
                    type = 1
                    tmp_string += "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } %.4f %.4f %.4f" % (
                                    atom2._atom_org_name,
                                    atom2._atom_type,
                                    atom2._res_original_id,
                                    atom2._res_type,
                                    atom2._chain,
                                    atom2._x,
                                    atom2._y,
                                    atom2._z)
                if atom2._sub_atom_type == 2:
                    # atom2 is a ligand atom
                    type = 1
                    tmp_string += "{ atomId=%i atmName=%s atmType=%s } %.4f %.4f %.4f" % (
                                    atom2.atom_org_id,
                                    atom2._atom_org_name,
                                    atom2._atom_type,
                                    atom2._x,
                                    atom2._y,
                                    atom2._z)
                if atom2._sub_atom_type == 3:
                    # atom2 is a water
                    type = 2
                    tmp_string += "{ hydLay=%i watType=%i } %.4f %.4f %.4f" % (
                                    atom2.hydration_layer,
                                    atom2._water_type,
                                    atom2._x,
                                    atom2._y,
                                    atom2._z)
                if atom2._sub_atom_type == 4:
                    # atom2 is a void
                    type = 3
                    tmp_string += "{ r=%.4f } %.4f %.4f %.4f" % (
                                    atom2._probe_radius,
                                    atom2._x,
                                    atom2._y,
                                    atom2._z)
            elif atom1._sub_atom_type == 2:
                # atom1 is a ligand atom
                tmp_string = "{ atmId=%i atmName=%s atmType=%s } P %.4f %.4f %.4f" % (
                                atom1.atom_org_id,
                                atom1._atom_org_name,
                                atom1._atom_type,
                                atom1._x,
                                atom1._y,
                                atom1._z)
                if atom2._sub_atom_type == 1:
                    # atom2 is a protein atom
                    type = 1
                    tmp_string += "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } %.4f %.4f %.4f" % (
                                    atom2._atom_org_name,
                                    atom2._atom_type,
                                    atom2._res_original_id,
                                    atom2._res_type,
                                    atom2._chain,
                                    atom2._x,
                                    atom2._y,
                                    atom2._z)
                if atom2._sub_atom_type == 2:
                    # atom2 is a ligand atom
                    type = 1
                    tmp_string += "{ atomId=%i atmName=%s atmType=%s } %.4f %.4f %.4f" % (
                                    atom2.atom_org_id,
                                    atom2._atom_org_name,
                                    atom2._atom_type,
                                    atom2._x,
                                    atom2._y,
                                    atom2._z)
                if atom2._sub_atom_type == 3:
                    # atom2 is a water
                    type = 2
                    tmp_string += "{ hydLay=%i watType=%i } %.4f %.4f %.4f" % (
                                    atom2.hydration_layer,
                                    atom2._water_type,
                                    atom2._x,
                                    atom2._y,
                                    atom2._z)
                if atom2._sub_atom_type == 4:
                    # atom2 is a void
                    type = 3
                    tmp_string += "{ r=%.4f } %.4f %.4f %.4f" % (
                                    atom2._probe_radius,
                                    atom2._x,
                                    atom2._y,
                                    atom2._z)

            dot_string = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
   
            if ( type == 1 and polarity == 0):
                polar_contacts.append(tmp_string)
                polar_contact_dots.append(dot_string)
            elif ( type == 1 and polarity == 2):
                apolar_contacts.append(tmp_string)
                apolar_contact_dots.append(dot_string)
            elif ( type == 2 and polarity == 0):
                polar_water_contacts.append(tmp_string)
                polar_water_contact_dots.append(dot_string)
            elif ( type == 2 and polarity == 2):
                apolar_water_contacts.append(tmp_string)
                apolar_water_contact_dots.append(dot_string)
            elif ( type == 3 and polarity == 0):
                polar_void_contacts.append(tmp_string)
                polar_void_contact_dots.append(dot_string)
            elif ( type == 3 and polarity == 2):
                apolar_void_contacts.append(tmp_string)
                apolar_void_contact_dots.append(dot_string)
        
        kin_list.extend(polar_contacts)
        kin_list.extend(apolar_contacts)
        kin_list.extend(polar_water_contacts)
        kin_list.extend(apolar_water_contacts)
        kin_list.extend(polar_void_contacts)
        kin_list.extend(apolar_void_contacts)
        kin_list.extend(polar_contact_dots)
        kin_list.extend(apolar_contact_dots)
        kin_list.extend(polar_water_contact_dots)
        kin_list.extend(apolar_water_contact_dots)
        kin_list.extend(polar_void_contact_dots)
        kin_list.extend(apolar_void_contact_dots)

        return "\n".join(kin_list)

    def get_residue_residue_contacts(self):
        """
        This method returns a text string containting a residue residue
        contact csv file.
        
        The format is outlined below:
        resid, resname, chain, resid, resname, chain, total no. of polar contacts, total no. of apolar contacts
        """
        # This can be accomplished by looping through all contacts give each
        # residue a unique name contact to dictionnary of dictionarry.
        lines = []
        res_res = {}
        contacts = self.interaction
        for contact in contacts:
            atom1 = contact._atom1
            atom2 = contact._atom2
            polarity = contact._contact_polarity
            atom1_res_id = get_res_id(atom1)
            atom2_res_id = get_res_id(atom2)
            if atom1._sub_atom_type == 1 or atom1._sub_atom_type == 2:
                res_res.setdefault(atom1_res_id, {}).setdefault(atom2_res_id, [0,0])
                if polarity == 0:
                    res_res[atom1_res_id][atom2_res_id][0] += 1
                if polarity == 2:
                    res_res[atom1_res_id][atom2_res_id][1] += 1
        res1_keys = res_res.keys()
        res1_keys.sort()
        line = '"resid1","resname1","chain1","resid2","resname2","chain2","polar contacts","apolar contacts"'
        lines.append(line)
        for res1_key in res1_keys:
            res2_keys = res_res[res1_key].keys()
            res2_keys.sort()
            for res2_key in res2_keys:
                polar_contacts = res_res[res1_key][res2_key][0]
                apolar_contacts = res_res[res1_key][res2_key][1]
                id1, res_id1, hetnam1, chain1 = res1_key.split('_') 
                id2, res_id2, hetnam2, chain2 = res2_key.split('_') 
                line = '%s,"%s","%s",%s,"%s","%s",%i,%i' % (res_id1,
                                                    hetnam1,
                                                    chain1,
                                                    res_id2,
                                                    hetnam2,
                                                    chain2,
                                                    polar_contacts,
                                                    apolar_contacts)
                lines.append(line)
        return '\n'.join(lines)


    def get_kin_binding_site_contacts(self):
        """function get_kin_binding_site_contacts

        This method returns a text sting. The text string can be
        written to a .kin file and visualised using mage.

        The kin text string contains:
        1) A description of the file (@text...)
        2) Dots and vectors representing the contacts in the binding site
        
        returns string

        """

        kin_list = []
        info_text = """@text
                    This file contains vectors representing:
                    1) Contacts from ligand to protein
                    2) Contacts from ligand to waters
                    3) Contacts from ligand to cavities
                    4) Contacts from protein to waters
                    5) Contacts from protein to cavities
                    as well as dots representing the midpoint of contacts.
                    The contacts are characterised as polar or apolar.
                    """
        kin_list.append(re.sub(r'\n\s+', '\n', info_text))
        kin_list.append("\n@kinemage\n")
        contacts = self.interaction

        l_to_p_pol_list = []
        l_to_p_apol_list = []
        l_to_w_pol_list = []
        l_to_w_apol_list = []
        l_to_v_pol_list = []
        l_to_v_apol_list = []
        p_to_w_pol_list = []
        p_to_w_apol_list = []
        p_to_v_pol_list = []
        p_to_v_apol_list = []
        p_to_v_apol_list = []
        
        l_to_p_pol_dot = []
        l_to_p_apol_dot = []
        l_to_w_pol_dot = []
        l_to_w_apol_dot = []
        l_to_v_pol_dot = []
        l_to_v_apol_dot = []
        p_to_w_pol_dot = []
        p_to_w_apol_dot = []
        p_to_v_pol_dot = []
        p_to_v_apol_dot = []
        p_to_v_apol_dot = []
        
        l_to_p_pol_list.append("@vectorlist {L-P Polar} color=red width=3")
        l_to_p_apol_list.append("@vectorlist {L-P Apolar} color=green width=3")
        l_to_w_pol_list.append("@vectorlist {L-W Polar} color=red width=1")
        l_to_w_apol_list.append("@vectorlist {L-W Apolar} color=green width=1")
        l_to_v_pol_list.append("@vectorlist {L-V Polar} off color=red width=1")
        l_to_v_apol_list.append("@vectorlist {L-V Apolar} off color=green width=1")
        p_to_w_pol_list.append("@vectorlist {P-W Polar} color=red width=1")
        p_to_w_apol_list.append("@vectorlist {P-W Apolar} color=green width=1")
        p_to_v_pol_list.append("@vectorlist {P-V Polar} off color=red width=1")
        p_to_v_apol_list.append("@vectorlist {P-V Apolar} off color=green width=1")
        
        l_to_p_pol_dot.append("@dotlist {L-P Polar} color=red width=3")
        l_to_p_apol_dot.append("@dotlist {L-P Apolar} color=green width=3")
        l_to_w_pol_dot.append("@dotlist {L-W Polar} color=red width=1")
        l_to_w_apol_dot.append("@dotlist {L-W Apolar} color=green width=1")
        l_to_v_pol_dot.append("@dotlist {L-V Polar} off color=red width=1")
        l_to_v_apol_dot.append("@dotlist {L-V Apolar} off color=green width=1")
        p_to_w_pol_dot.append("@dotlist {P-W Polar} color=red width=1")
        p_to_w_apol_dot.append("@dotlist {P-W Apolar} color=green width=1")
        p_to_v_pol_dot.append("@dotlist {P-V Polar} off color=red width=1")
        p_to_v_apol_dot.append("@dotlist {P-V Apolar} off color=green width=1")
        
        for contact in contacts:
            atom1 = contact._atom1
            atom2 = contact._atom2
            polarity = contact._contact_polarity
            if (atom1._sub_atom_type == 2
                and atom2._sub_atom_type == 1
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s } P %.4f %.4f %.4f { atmName=%s atmType=%s resId=%i resType=%s chain=%s } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._atom_org_name,
                            atom2._atom_type,
                            atom2._res_original_id,
                            atom2._res_type,
                            atom2._chain,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                l_to_p_pol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                l_to_p_pol_dot.append(tmp_kin)
                                      
                                      
                                      
                
            if (atom1._sub_atom_type == 2
                and atom2._sub_atom_type == 1
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s } P %.4f %.4f %.4f { atmName=%s atmType=%s resId=%i resType=%s chain=%s } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._atom_org_name,
                            atom2._atom_type,
                            atom2._res_original_id,
                            atom2._res_type,
                            atom2._chain,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                l_to_p_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                l_to_p_apol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 2
                and atom2._sub_atom_type == 3
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s } P %.4f %.4f %.4f { hydLay=%i watType=%i} %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2.hydration_layer,
                            atom2._water_type,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                l_to_w_pol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                l_to_w_pol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 2
                and atom2._sub_atom_type == 3
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s } P %.4f %.4f %.4f { hydLay=%i watType=%i} %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2.hydration_layer,
                            atom2._water_type,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                l_to_w_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                l_to_w_apol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 2
                and atom2._sub_atom_type == 4
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s } P %.4f %.4f %.4f { r=%.4f } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._probe_radius,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                l_to_v_pol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                l_to_v_pol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 2
                and atom2._sub_atom_type == 4
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s } P %.4f %.4f %.4f { r=%.4f } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._probe_radius,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                l_to_v_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                l_to_v_apol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom2._sub_atom_type == 3
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { hydLay=%i watType=%i} %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._hydration_layer,
                            atom2._water_type,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                p_to_w_pol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                p_to_w_pol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom2._sub_atom_type == 3
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { hydLay=%i watType=%i} %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._hydration_layer,
                            atom2._water_type,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                p_to_w_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                p_to_w_apol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom2._sub_atom_type == 4
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { r=%.4f } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._probe_radius,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                p_to_v_pol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                p_to_v_pol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom2._sub_atom_type == 4
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { r=%.4f } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._probe_radius,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                p_to_v_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                                        contact._x,
                                        contact._y,
                                        contact._z)
                p_to_v_apol_dot.append(tmp_kin)
                
        kin_list.extend(l_to_p_pol_list)
        kin_list.extend(l_to_p_apol_list)
        kin_list.extend(l_to_w_pol_list)
        kin_list.extend(l_to_w_apol_list)
        kin_list.extend(l_to_v_pol_list)
        kin_list.extend(l_to_v_apol_list)
        kin_list.extend(p_to_w_pol_list)
        kin_list.extend(p_to_w_apol_list)
        kin_list.extend(p_to_v_pol_list)
        kin_list.extend(p_to_v_apol_list)
        
        kin_list.extend(l_to_p_pol_dot)
        kin_list.extend(l_to_p_apol_dot)
        kin_list.extend(l_to_w_pol_dot)
        kin_list.extend(l_to_w_apol_dot)
        kin_list.extend(l_to_v_pol_dot)
        kin_list.extend(l_to_v_apol_dot)
        kin_list.extend(p_to_w_pol_dot)
        kin_list.extend(p_to_w_apol_dot)
        kin_list.extend(p_to_v_pol_dot)
        kin_list.extend(p_to_v_apol_dot)

        return "\n".join(kin_list)

    def get_kin_interface_contacts(self):
        """function get_kin_interface_contacts
        
        This method returns a text sting. The text string can be
        written to a .kin file and visualised using mage.

        The kin text string contains:
        1) A description of the file (@text...)
        2) Dots and vectors representing the contacts in the interface
        
        returns string

        """
        
        kin_list = []
        info_text = """@text
                    This file contains vectors representing:
                    1) Contacts from subunit1 to subunit2
                    2) Contacts from subunit1 to waters
                    3) Contacts from subunit1 to probes
                    4) Contacts from subunit2 to waters
                    5) Contacts from subunit2 to probes
                    as well as dots representing the midpoint of contacts.
                    The contacts are characterised as polar or apolar.
                    """
        kin_list.append(re.sub(r'\n\s+', '\n', info_text))
        kin_list.append("@kinemage")
        contacts = self.interaction

        s1_to_s2_pol_list = []
        s1_to_s2_apol_list = []
        s1_to_wat_pol_list = []
        s1_to_wat_apol_list = []
        s1_to_void_pol_list = []
        s1_to_void_apol_list = []
        s2_to_wat_pol_list = []
        s2_to_wat_apol_list = []
        s2_to_void_pol_list = []
        s2_to_void_apol_list = []

        s1_to_s2_pol_dot = []
        s1_to_s2_apol_dot = []
        s1_to_wat_pol_dot = []
        s1_to_wat_apol_dot = []
        s1_to_void_pol_dot = []
        s1_to_void_apol_dot = []
        s2_to_wat_pol_dot = []
        s2_to_wat_apol_dot = []
        s2_to_void_pol_dot = []
        s2_to_void_apol_dot = []
        
        s1_to_s2_pol_list.append("@vectorlist {S1-S2 Polar} color=red width=3")
        s1_to_s2_apol_list.append("@vectorlist {S1-S2 Apolar}  color=green width=3")
        s1_to_wat_pol_list.append("@vectorlist {S1-W Polar} color=red width=1")
        s1_to_wat_apol_list.append("@vectorlist {S1-W Apolar} color=green width=1")
        s1_to_void_pol_list.append("@vectorlist {S1-V Polar} off color=red width=1")
        s1_to_void_apol_list.append("@vectorlist {S1-V Apolar} off color=green width=1")
        s2_to_wat_pol_list.append("@vectorlist {S2-W Polar} color=red width=1")
        s2_to_wat_apol_list.append("@vectorlist {S2-W Apolar} color=green width=1")
        s2_to_void_pol_list.append("@vectorlist {S2-V Polar} off color=red width=1")
        s2_to_void_apol_list.append("@vectorlist {S2-V Apolar} off color=green width=1")

        s1_to_s2_pol_dot.append("@dotlist {S1-S2 Polar} color=red width=3")
        s1_to_s2_apol_dot.append("@dotlist {S1-S2 Apolar} color=green width=3")
        s1_to_wat_pol_dot.append("@dotlist {S1-W Polar} color=red width=1")
        s1_to_wat_apol_dot.append("@dotlist {S1-W Apolar} color=green width=1")
        s1_to_void_pol_dot.append("@dotlist {S1-V Polar} off color=red width=1")
        s1_to_void_apol_dot.append("@dotlist {S1-V Apolar} off color=green width=1")
        s2_to_wat_pol_dot.append("@dotlist {S2-W Polar} color=red width=1")
        s2_to_wat_apol_dot.append("@dotlist {S2-W Apolar} color=green width=1")
        s2_to_void_pol_dot.append("@dotlist {S2-V Polar} off color=red width=1")
        s2_to_void_apol_dot.append("@dotlist {S2-V Apolar} off color=green width=1")
        
        for contact in contacts:
            atom1 = contact._atom1
            atom2 = contact._atom2
            polarity = contact._contact_polarity
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 1
                and atom2._sub_atom_type == 1
                and atom2._molecule_instance == 2
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { atomName=%s atmType=%s resId=%i resType=%s chain=%s } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._atom_org_name,
                            atom2._atom_type,
                            atom2._res_original_id,
                            atom2._res_type,
                            atom2._chain,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s1_to_s2_pol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s1_to_s2_pol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 1
                and atom2._sub_atom_type == 1
                and atom2._molecule_instance == 2
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { atomName=%s atmType=%s resId=%i resType=%s chain=%s } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._atom_org_name,
                            atom2._atom_type,
                            atom2._res_original_id,
                            atom2._res_type,
                            atom2.chain,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s1_to_s2_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s1_to_s2_apol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 1
                and atom2._sub_atom_type == 3
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { hydLay=%i watType=%i } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._hydration_layer,
                            atom2._water_type,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s1_to_wat_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s1_to_wat_pol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 1
                and atom2._sub_atom_type == 3
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { hydLay=%i watType=%i } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._hydration_layer,
                            atom2._water_type,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s1_to_wat_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s1_to_wat_apol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 1
                and atom2._sub_atom_type == 4
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { r=%.4f } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._probe_radius,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s1_to_void_pol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s1_to_void_pol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 1
                and atom2._sub_atom_type == 4
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { r=%.4f } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._probe_radius,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s1_to_void_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s1_to_void_apol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 2
                and atom2._sub_atom_type == 3
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { hydLay=%i watType=%i } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._hydration_layer,
                            atom2._water_type,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s2_to_wat_pol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s2_to_wat_pol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 2
                and atom2._sub_atom_type == 3
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { hydLay=%i watType=%i } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._hydration_layer,
                            atom2._water_type,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s2_to_wat_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s2_to_wat_apol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 2
                and atom2._sub_atom_type == 4
                and polarity == 0):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { r=%.4f } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._probe_radius,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s2_to_void_pol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s2_to_void_pol_dot.append(tmp_kin)
                
            if (atom1._sub_atom_type == 1
                and atom1._molecule_instance == 2
                and atom2._sub_atom_type == 4
                and polarity == 2):
                tmp_kin = "{ atmName=%s atmType=%s resId=%i resType=%s chain=%s } P %.4f %.4f %.4f { r=%.4f } %.4f %.4f %.4f" % (
                            atom1._atom_org_name,
                            atom1._atom_type,
                            atom1._res_original_id,
                            atom1._res_type,
                            atom1._chain,
                            atom1._x,
                            atom1._y,
                            atom1._z,
                            atom2._probe_radius,
                            atom2._x,
                            atom2._y,
                            atom2._z)
                s2_to_void_apol_list.append(tmp_kin)

                tmp_kin = "{} %.4f %.4f %.4f" % (
                            contact._x,
                            contact._y,
                            contact._z)
                s2_to_void_apol_dot.append(tmp_kin)
        
        kin_list.extend(s1_to_s2_pol_list)
        kin_list.extend(s1_to_s2_apol_list)
        kin_list.extend(s1_to_wat_pol_list)
        kin_list.extend(s1_to_wat_apol_list)
        kin_list.extend(s1_to_void_pol_list)
        kin_list.extend(s1_to_void_apol_list)
        kin_list.extend(s2_to_wat_pol_list)
        kin_list.extend(s2_to_wat_apol_list)
        kin_list.extend(s2_to_void_pol_list)
        kin_list.extend(s2_to_void_apol_list)
        
        kin_list.extend(s1_to_s2_pol_dot)
        kin_list.extend(s1_to_s2_apol_dot)
        kin_list.extend(s1_to_wat_pol_dot)
        kin_list.extend(s1_to_wat_apol_dot)
        kin_list.extend(s1_to_void_pol_dot)
        kin_list.extend(s1_to_void_apol_dot)
        kin_list.extend(s2_to_wat_pol_dot)
        kin_list.extend(s2_to_wat_apol_dot)
        kin_list.extend(s2_to_void_pol_dot)
        kin_list.extend(s2_to_void_apol_dot)

        return '\n'.join(kin_list)
        
    def get_number_of_atoms(self):
        macro_mol_atoms1 = 0
        macro_mol_atoms2 = 0
        ligand_atoms = 0
        for atom in self.macro_molecule:
            if atom._include == 1:
                if atom._molecule_instance == 1:
                    macro_mol_atoms1 = macro_mol_atoms1 + 1
                elif atom._molecule_instance == 2:
                    macro_mol_atoms2 = macro_mol_atoms2 + 1
        ligand_atoms = len(self.ligand)
        atoms_string = """Number of protein 1 atoms              : %i
                          Number of protein 2 atoms              : %i
                          Number of ligand atoms                 : %i""" % (
                             macro_mol_atoms1, macro_mol_atoms2, ligand_atoms)

        return re.sub(r'\n\s+', '\n', atoms_string)
        
    def get_number_of_atoms_csv(self):
        macro_mol_atoms1 = 0
        macro_mol_atoms2 = 0
        ligand_atoms = 0
        for atom in self.macro_molecule:
            if atom._include == 1:
                if atom._molecule_instance == 1:
                    macro_mol_atoms1 = macro_mol_atoms1 + 1
                elif atom._molecule_instance == 2:
                    macro_mol_atoms2 = macro_mol_atoms2 + 1
        ligand_atoms = len(self.ligand)
        atoms_csv = "%i,%i,%i," % (macro_mol_atoms1,
                                   macro_mol_atoms2,
                                   ligand_atoms)
        return atoms_csv
        
# Test Assembly class with 'python Assembly.py'
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    import unittest
    suite = doctest.DocFileSuite('doc_tests/test_Assembly.txt',
        optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)
    unittest.TextTestRunner().run(suite)
