from Point import Point
from Atom import Atom

class Contact(Point):
    """Class Contact

    The class Contact inherits all the attibutes and methods of the Point
    class.  The coordinates of Contact refer to the midpoint in between the
    two atoms of interest.

    The Contact class has four further attributes:
    1) __atom1 (reference to one of the atoms in contact)
    2) __atom2 (reference to the other atom in contact)
    3) _contact_polarity (a contact can be polar (0) or apolar (2))
    4) __d2 (the distance squared between atom1 and atom2)
    5) _contact_complimentarity
        (one polar atom and one apolar atom -> not complimentary (0)
        both polar or both apolar atoms     -> complimentary (1) )

    In looping through an atomList setting all the possible contacts between
    all the pairs of atoms within a certain distance it should be noted that
    every unique contact will be entered twice into the
    ContactList.__contactList.
    1) atom1-atom2
    2) atom2-atom1

    """
    

    def __init__(self, a1, a2, polarity):
        """Initialise the contact
        
        a1: Atom reference
        a2: Atom reference
        polarity: int

        This method sets all the attributes of the Contact class.
        
        """
        self._atom1 = a1
        self._atom2 = a2
        self._contact_polarity = polarity
        distance2 = a1.get_distance2(a2)
        self._distance2 = distance2
        # set the midpoint
        tmp_point = (a1  + a2) / 2
        Point.__init__(self, tmp_point.x, tmp_point.y, tmp_point.z)
        # set the complimentarity
        if a2._sub_atom_type != 4:
            if a1.atom_polarity == a2.atom_polarity:
                self._contact_complimentarity = 1
            else:
                self._contact_complimentarity = 0
        else:
            self._contact_complimentarity = 1
    
    @property
    def atom1(self):
        """
        Return atom1.
        """
        return self._atom1
        
    @property
    def atom2(self):
        """
        Return atom2.
        """
        return self._atom2
        
    @property
    def contact_polarity(self):
        """
        Return contact polarity.
        """
        return self._contact_polarity
    
    @property
    def distance2(self):
        """
        Return disstance squared between atom1 and atom2.
        """
        return self._distance2
    
    @property
    def contact_complimentarity(self):
        """
        Return contact complimentarity.
        """
        return self._contact_complimentarity

    def clear(self):
        """
        Remove atom references.
        """
        del self._atom1
        del self._atom2
        
# Test contact by 'python Contact.py'
if __name__ == "__main__":
    import doctest
    import unittest
    suite = doctest.DocFileSuite('doc_tests/test_Contact.txt',
        optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)
    unittest.TextTestRunner().run(suite)
