import sys
import string
import re
from Bio.PDB.PDBParser import PDBParser
from mol2 import Mol2
from Point import Point
from Assembly import Assembly
from UnitSurfacePoints import UnitSurfacePoints
from ContactList import ContactList

IOError = "io error"

class IO:
    """Class IO

The class IO deals with the input and output of PRO_ACT2.
The attributes of the IO class refer to the user input of filenames and
options for running PRO_ACT2. These are read in using the method
readCommandLine().

The PDB files are read in using the BioPython PDB module.

The Mol2 files are read in using the rough mol2 class, which isn't beautiful,
but it does the job.
    """
    # Attributes:
    __fileName1 = None  # (string) 
    __fileType1 = None  # (string) 
    __residueSelection1 = None  # (list) 
    __chainSelection1 = None  # (list) 
    __includeExpWaters1 = False  # (boolean) 
    __fileName2 = None  # (string) 
    __fileType2 = None  # (string) 
    __residueSelection2 = None  # (list) 
    __chainSelection2 = None  # (list) 
    __includeExpWaters2 = False  # (boolean) 
    __onlyBindingSite = False  # (boolean) 
    __onlyInterface = False  # (boolean) 
    __onlyReturnInteraction = False  # (boolean) 
    __numberOfRuns = 1  # (int) 
    __numberOfHydrationLayers = 1  # (int) 
    __strictWaterType = True
    
    # Operations
    def readPdb(self, assembly, moleculeInstance, filename, residueSelection,
        chainSelection, includeExpWaters):
        """function readPdb
        
        assembly: Assembly
        moleucleInstance: int
        filename: string
        residueSelection: list
        chainSelection: list
        includeExpWaters: boolean

        This method reads in all the atoms of the PDB file specified by the
        residueSelection and the chainSelection. They should have the
        following form:
        1) chainSelection= 'A' (chain A) or 'AB' (chain A and chain B)
        2) residueSelection= '3' (residue 3) or '3-5' (residues 3,4 and 5)
                             or '3-5,10' (residues 3,4,5 and 10)

        If the includeExpWaters is set to True the method will read the
        experimental waters (of the selected chains/residues) into the
        Hydration class.

        In cases of diorder the Bio.PDB module automatically selects the atom
        with the highest occupancy.

        Protein atoms will be read into the MacroMolecule class.

        All hetatoms are currently excluded.

        Nucleic acids are currently not supported.
        
        returns boolean
        """
        M = assembly.getMacroMolecule()
        if includeExpWaters:
            H = assembly.getHydration()
        
        # This is a hack to overcome the fact that the pdb parser
        # writes it's errors to stdout
        class FileFaker:
            def write(self, string):
                sys.stderr.write(string)
        import sys
        sys.stdout = FileFaker()
        
        p = PDBParser(PERMISSIVE=1)  # load the pdb parser object
        structure_id = filename.replace('.pdb', '')
        structure = p.get_structure(structure_id, filename)

        # Remove the hack
        sys.stdout = sys.__stdout__

        model = structure[0]

        for chain in model.get_list():
            if chainSelection == 'all':
                useChain = True
            else:
                useChain = False
                for c in chainSelection:
                    if chain.get_id() == c:
                        useChain = True
                        break
            if not(useChain):  # do not use this chain
                continue       # next chain
            for residue in chain.get_list():
                if residueSelection == 'all':
                    useRes = True
                else:
                    useRes = False
                    for r in residueSelection:
                        if residue.get_id()[1] == r:
                            useRes = True
                            break
                if includeExpWaters == False:
                    if residue.get_id()[0] == 'W':  # res is a water
                        useRes = False
                if string.find(residue.get_id()[0], 'H_') != -1:  # hetatm
                    useRes = False                                # exclude
                if not(useRes):  # do not use this residue
                    continue     # next residue
                for atom in residue.get_list():

                    # Get atom type
                    atomType = atom.get_fullname()[1]

                    # Mild sanity check of atom names
                    if atomType == 'H':   # hydrogen exclude
                        continue
                    if re.match('H', atom.get_fullname()) != None:
                        continue          # exclude weird hydrogens
#                   if re.match('LP', atom.get_fullname()) != None:
#                       continue          # exclude weird disorder
                        
                    if ( atomType != 'C'
                        and atomType != 'N'
                        and atomType != 'O'
                        and atomType != 'S'
                        and atomType != 'H'):
                        type = atom.get_fullname()
                        errorMessage = "Atom type?:"+type+residue.get_id()[0]
                        raise IOError, errorMessage
                    if residue.get_id()[0] == 'W':  # res is a water
                        H.atomInput(
                            residue.get_id()[1],   # residueId
                            residue.get_resname(), # resType
                            chain.get_id(),        # chain
                            None,                  # Can't get the atom id
                            atom.get_fullname(),   # atomName
                            atomType,
                            atom.get_coord()[0],   # x
                            atom.get_coord()[1],   # y
                            atom.get_coord()[2],   # z
                            3,                     # subAtomType
                            moleculeInstance
                        )
                    else:
                        M.atomInput(
                            residue.get_id()[1],   # residueId
                            residue.get_resname(), # resType
                            chain.get_id(),        # chain
                            None,                  # Can't get the atom id
                            atom.get_fullname(),   # atomName
                            atomType,
                            atom.get_coord()[0],   # x
                            atom.get_coord()[1],   # y
                            atom.get_coord()[2],   # z
                            1,                     # subAtomType
                            moleculeInstance
                        )

        return True
    
    def readMol2(self, assembly, moleculeInstance, filename):
        """function readMol2
        
        assembly: Assembly
        moleulceInstance: int
        filename: string

        This method reads the mol2 file into the Ligand class.
        
        returns boolean
        """
        L = assembly.getLigand()
        ligand = Mol2()
        ligand.readMol2(filename)
        for i in range(len(ligand.atomRec)):
            if ligand.getAtomType(i) == 'H':
                continue
            L.atomInput(
                ligand.atomRec[i]['atom_id'],
                ligand.atomRec[i]['atom_name'],
                ligand.getAtomType(i),
                ligand.atomRec[i]['x'],
                ligand.atomRec[i]['y'],
                ligand.atomRec[i]['z'],
                2,
                moleculeInstance)
        return True
    
    def readCommandLine(self):
        """function readCommandLine

        This method is used to read the user input from the command line. This
        method does most of the quality control checking.
        Some common usage examples are listed below.
        
        Typical command line input:

        Simple
        main.py fileName1='1hsg.pdb' fileType1='pdb'

        Protein-Ligand interaction
        main.py fileName1='1hsg.pdb' fileType1='pdb' filename2='ligand.mol2'
        fileType2='mol2' onlyBindingSite=True

        Dimer interface
        main.py fileName1='1hsg.pdb' fileType1='pdb' chainSelection1='A'
            fileName2='1hsg.pdb' fileType2='pdb' chainSelection2='B'
            onlyInterface=True

        Protein-Ligand thermodynamics
        main.py fileName1='1hsg.pdb' fileType1='pdb' filename2='ligand.mol2'
            fileType2='mol2' onlyBindingSite=True onlyReturnInteraction=True
            numberOfHydrationLayers=2 numberOfRuns=10

        returns boolean
        """
        inputDict = {
            'fileName1'                 : None,
            'fileType1'                 : None,
            'chainSelection1'           : 'all',
            'residueSelection1'         : 'all',
            'includeExpWaters1': False,
            'fileName2'                 : None,
            'fileType2'                 : None,
            'chainSelection2'           : 'all',
            'residueSelection2'         : 'all',
            'includeExpWaters2': False,
            'onlyBindingSite'           : False,
            'onlyInterface'             : False,
            'onlyReturnInteraction'     : False,
            'numberOfRuns'              : 1,
            'numberOfHydrationLayers'   : 1,
            'strictWaterType'           : True,
        }
        commandLine = sys.argv
        for i, command in enumerate(commandLine):
            if i == 0:     # script name
                continue
            pair = string.split(command, '=')
            if pair[1] == 'True':
                inputDict[pair[0]] = True
            elif pair[1] == 'False':
                inputDict[pair[0]] = False
            elif pair[1] == '':
                raise IOError, 'command line input ' + pair[0] + "= ?"
            else:
                inputDict[pair[0]] = pair[1]

        # Quality control
        fileCount = 1
        if inputDict['fileName1'] == None:
            raise IOError, "no structure input file use: fileName1='YourFile'"
        if inputDict['fileType1'] == None:
            raise IOError, "no filetype for file1 use: fileType1='pdb/mol2'"
        if inputDict['fileName2'] != None and inputDict['fileType2'] == None:
            raise IOError, "no filetype for file2 use: fileType2='pdb/mol2'"
        if inputDict['fileName2'] != None:
            fileCount = 2
        if fileCount != 2:
            if inputDict['onlyBindingSite'] == True:
                raise IOError, "onlyBindingSite=True requires two struc. files"
            if inputDict['onlyInterface'] == True:
                raise IOError, "onlyInterface=True requires two struc. files"
            if inputDict['onlyReturnInteraction'] == True:
                raise IOError, "onlyReturnInteraction requires two struc. files"
        if inputDict['onlyBindingSite'] == True:
            if inputDict['fileType1'] != 'pdb':
                raise IOError, "onlyBindingSite requires file1 to be 'pdb'"
            if inputDict['fileType2'] != 'mol2':
                raise IOError, "onlyBindingSite requires file2 to be 'mol2'"
        if inputDict['onlyInterface'] == True:
            if inputDict['fileType1'] != 'pdb':
                raise IOError, "onlyBindingSite requires file1 to be 'pdb'"
            if inputDict['fileType2'] != 'pdb':
                raise IOError, "onlyBindingSite requires file2 to be 'pdb'"

        # Convert the chain selections into lists
        def chainList(x):
            if x == 'all':
                return 'all'
            x = string.strip(x)  # remove preceeding and trailing whitespaces
            x = string.upper(x)  # upper case
            list = []            # make list
            for letter in x:
                list.append(letter)
            return list

        chainSel1 = chainList(inputDict['chainSelection1'])
        chainSel2 = chainList(inputDict['chainSelection2'])

        # Convert the residue selecitons into lists
        def resList(x):
            if x == 'all':
                return 'all'
            x = string.strip(x)  # remove preceeding and trailing whitespaces
            xlist = string.split(x, ',')  # split selections
            list = []
            for selection in xlist:
                if string.find(selection, ':') == -1:  # not a range
                    resId = string.atoi(selection)     # convert to int
                    list.append(resId)                 # append to list
                else:                                       # a range
                    indexes = string.split(selection, ':')  # get the indices
                    start = string.atoi(indexes[0])         # convert to int
                    stop = string.atoi(indexes[1])          # convert to int
                    stop = stop + 1                         # be inclusive
                    for i in range(start, stop):
                        list.append(i)
            return list
        
        resSel1 = resList(inputDict['residueSelection1'])
        resSel2 = resList(inputDict['residueSelection2'])

        # Set the variables
        self.__fileName1 = inputDict['fileName1']
        self.__fileType1 = inputDict['fileType1']
        self.__residueSelection1 = resSel1
        self.__chainSelection1 = chainSel1
        self.__includeExpWaters1 = inputDict['includeExpWaters1']
        self.__fileName2 = inputDict['fileName2']
        self.__fileType2 = inputDict['fileType2']
        self.__residueSelection2 = resSel2
        self.__chainSelection2 = chainSel2
        self.__includeExpWaters2 = inputDict['includeExpWaters2']
        self.__onlyBindingSite = inputDict['onlyBindingSite']
        self.__onlyInterface = inputDict['onlyInterface']
        self.__onlyReturnInteraction = inputDict['onlyReturnInteraction']
        self.__numberOfRuns = inputDict['numberOfRuns']
        self.__numberOfHydrationLayers = inputDict['numberOfHydrationLayers']
        self.__strictWaterType = inputDict['strictWaterType']

        return True
    
    def inputInfoString(self):
        """function inputInfoString
        
        This function writes out the values of the attributes of the IO class.

        returns string
        """
        s =  "------------------------------------------------------\n"
        s += "           Structure File 1\n"
        s += "File                       :" + str(self.__fileName1)
        s += "\nType                       :" + str(self.__fileType1)
        s += "\nChain Selection            :" + str(self.__chainSelection1)
        s += "\nResidue Selection          :" + str(self.__residueSelection1)
        s += "\nInclude Experimental Waters:" + str(self.__includeExpWaters1)
        s += "\n------------------------------------------------------\n"
        s += "           Structure File 2\n"
        s += "File                       :" + str(self.__fileName2)
        s += "\nType                       :" + str(self.__fileType2)
        s += "\nChain Selection            :" + str(self.__chainSelection2)
        s += "\nResidue Selection          :" + str(self.__residueSelection2)
        s += "\nInclude Experimental Waters:" + str(self.__includeExpWaters2)
        s += "\n------------------------------------------------------\n"
        s += "                    Options\n"
        s += "Only Binding Site          :" + str(self.__onlyBindingSite)
        s += "\nOnly Interface             :" + str(self.__onlyInterface)
        s += "\nOnly Return Interaction    :"
        s += str(self.__onlyReturnInteraction)
        s += "\nNumber Of Runs             :" + str(self.__numberOfRuns)
        s += "\nNumber Of Hydration Layers :"
        s += str(self.__numberOfHydrationLayers)
        s += "\nStrict Water Type          :" + str(self.__strictWaterType)
        s += "\n------------------------------------------------------\n"

        return s
    
    def getOnlyBindingSite(self):
        """function getOnlyBindingSite

        This method returns the __onlyBindingSite attribute.
        
        returns boolean
        """
        return self.__onlyBindingSite
    
    def getOnlyInterface(self):
        """function getOnlyInterface
        
        This method returns the __onlyInterface attribute.
        
        returns boolean
        """
        return self.__onlyInterface
    
    def getOnlyReturnInteraction(self):
        """function getOnlyReturnInteraction
        
        This method returns the __onlyReturnInteraction attribute.
        
        returns boolean
        """
        return self.__onlyReturnInteraction
    
    def getNumberOfRuns(self):
        """function getNumberOfRuns
        
        This method returns the __numberOfRuns attribute.

        returns int
        """
        return self.__numberOfRuns
    
    def getNumberOfHydrationLayers(self):
        """function getNumberOfHydrationLayers
        
        This method returns the __numberOfHydrationLayers attribute.

        returns int
        """
        return self.__numberOfHydrationLayers
    
    def getFileName1(self):
        """function getFileName1
        
        This method returns the __fileName1 attribute.

        returns string
        """
        return self.__fileName1
    
    def getFileType1(self):
        """function getFileType1
        
        This method returns the __fileType1 attribute.

        returns string
        """
        return self.__fileType1
    
    def getResidueSelection1(self):
        """function getResidueSelection1
        
        This method returns the __residueSelection1 attribute.

        returns list
        """
        return self.__residueSelection1
    
    def getChainSelection1(self):
        """function getChainSelection1
        
        This method returns the __chainSelection1 attribute.

        returns list
        """
        return self.__chainSelection1
    
    def getIncludeExpWaters1(self):
        """function getIncludeExpWaters1
        
        This method returns the __includeExpWaters1 attribute.

        returns boolean
        """
        return self.__includeExpWaters1
    
    def getFileName2(self):
        """function getFileName2
        
        This method returns the __fileName2 attribute.

        returns string
        """
        return self.__fileName2
    
    def getFileType2(self):
        """function getFileType2
        
        This method returns the __fileType2 attribute.

        returns string
        """
        return self.__fileType2
    
    def getResidueSelection2(self):
        """function getResidueSelection2
        
        This method returns the __residueSelection2 attribute.

        returns list
        """
        return self.__residueSelection2
    
    def getChainSelection2(self):
        """function getChainSelection2
        
        This method returns the __chainSelection2 attribute.

        returns list
        """
        return self.__chainSelection2
    
    def getIncludeExpWaters2(self):
        """function getIncludeExpWaters2
        
        This method returns the __includeExpWaters2 attribute.

        returns boolean
        """
        return self.__includeExpWaters2
    
    def getStrictWaterType(self):
        """function getStrictWaterType
        
        This method returns the __strictWaterType attribute.

        returns boolean
        """
        return self.__strictWaterType
    
    
# Testing IO class
if __name__ == "__main__":
    io = IO()
    print io.inputInfoString()
    print "Testing simple functions..."
    print "Get only binding site:", io.getOnlyBindingSite()
    print "Get only interface   :", io.getOnlyInterface()
    print "Get only ret interact:", io.getOnlyReturnInteraction()
    print "Get number of runs   :", io.getNumberOfRuns()
    print "Get number of hyd lay:", io.getNumberOfHydrationLayers()
    print "FileName1            :", io.getFileName1()
    print "FileName2            :", io.getFileName2()
    print "FileType1            :", io.getFileType1()
    print "FileType2            :", io.getFileType2()
    print "Get res selectio1    :", io.getResidueSelection1()
    print "Get res selectio2    :", io.getResidueSelection2()
    print "Get chain selection1 :", io.getChainSelection1()
    print "Get chain selection2 :", io.getChainSelection2()
    print "Get include wat1     :", io.getIncludeExpWaters1()
    print "Get include wat2     :", io.getIncludeExpWaters2()
    print "Get strict wat type  :", io.getStrictWaterType()
