class Point(object):
    """Class Point

    A point in three dimensional space defined by the coordinates x, y and z. A
    point can determine its distance from another point.

    Attributes:
    x -- x coordiante (float)
    y -- y coordiante (float)
    z -- z coordiante (float)

    Methods:
    __init__      -- Initialize an instance.
    __del__       -- Delete an instance.
    __add__       -- Vector addition.
    __sub__       -- Vector subtraction.
    __mul__       -- Vector multiplication.
    __div__       -- Vector division.
    __str__       -- String representation.
    get_distance2 -- Return the squared distance to another point.
    clashes       -- Check if another point is within a cutoff distance.

    Examples:
    >>> from Point import Point
    >>> p1 = Point(0.0, 0.0, 0.0)
    >>> p2 = Point(1.0, 2.0, 3.0)
    >>> p1.get_distance2(p2)
    14.0

    >>> p1.clashes(p2, 3)
    False
    >>> p1.clashes(p2, 7)
    True

    >>> p3 = p2 + p2
    >>> print p3
    <2.00, 4.00, 6.00>
    >>> p4 = p1 + 3.0
    >>> print p4
    <3.00, 3.00, 3.00>

    >>> p5 = p3 - p4
    >>> print p5
    <-1.00, 1.00, 3.00>
    >>> p6 = p5 - 3.0
    >>> print p6
    <-4.00, -2.00, 0.00>

    >>> p7 = p4 * p4
    >>> print p7
    <9.00, 9.00, 9.00>
    >>> p8 = p3 * -1
    >>> print p8
    <-2.00, -4.00, -6.00>

    >>> p9 = p4 / p4
    >>> print p9
    <1.00, 1.00, 1.00>
    >>> p10 = p3 / -0.5
    >>> print p10
    <-4.00, -8.00, -12.00>

    """
    # Slots saves memory if there are millions of instances of an object. The
    # line below can be commented out to get more flexibility out of this
    # class.
    __slots__ = '_x', '_y', '_z'

    def __init__(self, x, y, z):
        self._x = float(x)
        self._y = float(y)
        self._z = float(z)

    def __add__(self, other):
        """Function __add__

        This mehtod allows the coordinates of points to be added together, and
        for single values to be added to all coordinates.

        return Point
        """
        x = 0.0
        y = 0.0
        z = 0.0
        if issubclass(other.__class__,Point):
            x = self._x + other._x
            y = self._y + other._y
            z = self._z + other._z
        else:
            x = self._x + float(other)
            y = self._y + float(other)
            z = self._z + float(other)
        p = Point(x, y, z)
        return p

    def __sub__(self, other):
        """Function __sub__

        This mehtod allows the coordinates of points to be subtracted from
        each other, and for single values to be subtracted from all
        coordinates.

        return Point
        """
        x = 0.0
        y = 0.0
        z = 0.0
        if issubclass(other.__class__, Point):
            x = self._x - other._x
            y = self._y - other._y
            z = self._z - other._z
        else:
            x = self._x - float(other)
            y = self._y - float(other)
            z = self._z - float(other)
        p = Point(x, y, z)
        return p

    def __mul__(self, other):
        """Function __mul__

        This mehtod allows the coordinates of points to be multiplied together, and
        for single values multiply all coordinates.

        return Point
        """
        x = 0.0
        y = 0.0
        z = 0.0
        if issubclass(other.__class__, Point):
            x = self._x * other._x
            y = self._y * other._y
            z = self._z * other._z
        else:
            x = self._x * float(other)
            y = self._y * float(other)
            z = self._z * float(other)
        p = Point(x, y, z)
        return p

    def __div__(self, other):
        """Function __div__

        This mehtod allows the coordinates of points to be divided from each
        other, and for single values to be divided from all coordinates.

        return Point
        """
        x = 0.0
        y = 0.0
        z = 0.0
        if issubclass(other.__class__, Point):
            x = self._x / other._x
            y = self._y / other._y
            z = self._z / other._z
        else:
            x = self._x / float(other)
            y = self._y / float(other)
            z = self._z / float(other)
        p = Point(x, y, z)
        return p

    def __str__(self):
        """Function __str__

        This mehtod allows the coordinates of points to be easily viewed.

        return String
        """
        s = '<%.2f, %.2f, %.2f>' % (self._x, self._y, self._z)
        return s

    def get_x(self):
        """Get x cooridinate."""
        return self._x
    
    def set_x(self, x):
        """Set x cooridinate."""
        self._x = float(x)
    
    x = property(get_x, set_x)
    
    def get_y(self):
        """Get y cooridinate."""
        return self._y
    
    def set_y(self, y):
        """Set y cooridinate."""
        self._y = float(y)
    
    y = property(get_y, set_y)
    
    def get_z(self):
        """Get z cooridinate."""
        return self._z
    
    def set_z(self, z):
        """Set z cooridinate."""
        self._z = float(z)
    
    z = property(get_z, set_z)

    def get_distance2(self, p):
        """
        This method gets the distance squared to the Point p.
        
        p: Point
        
        returns distance2

        """
        # Original
#       dx = self._x - p._x
#       dy = self._y - p._y
#       dz = self._z - p._z
#       d2 = dx*dx + dy*dy + dz*dz
#       return d2

        # Optimized
        dx = self._x - p._x
        dy = self._y - p._y
        dz = self._z - p._z
        d2 = dx*dx + dy*dy + dz*dz
        return d2
    
    def clashes(self, target, cutoff):
        """
        This method tests if the Point target is within a the cutoff.
        
        target: Point
        cutoff: float

        returns True if the points clash

        """
        # Original
#       d2 = self.get_distance2(target)
#       c2 = cutoff*cutoff - 0.00001
#       if d2 < c2:
#           return True
#       else:
#           return False

        # Optimized
        c2 = cutoff*cutoff - 0.00001
        dx = self._x - target._x
        dx2 = dx*dx
        if dx2 > c2:
            return False
        dy = self._y - target._y
        dy2 = dy*dy
        if dy2 + dx2 > c2:
            return False
        dz = self._z - target._z
        dz2 = dz*dz
        if dz2 + dy2 + dx2  > c2:
            return False
        return True
   

# Test the class by running the command 'python Point.py' in the terminal
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    import unittest
    suite = doctest.DocFileSuite('doc_tests/test_Point.txt',
        optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)
    unittest.TextTestRunner().run(suite)
