from ProAct2Error import ProAct2Error

class Radii(object):
    """
    This class holds the different radii sets. The attribute radii_set
    controls which radii should be used.

    The point of the radii_set attribute is to make sure that all instances of
    the Radii class have the same state with regards to the radii set used.

    Examples:
    >>> from Radii import Radii
    >>> r1 = Radii()
    >>> print '%.2f' % r1._WILLIAMS_POLAR_CLOSE_CONST  
    1.50
    >>> 
    >>> print '%.2f' % r1._LI_AND_NUSSINOV_POLAR_CLOSE_CONST  
    1.37
    >>> r1.radii_set = 'WILLIAMS'
    >>> print '%.2f' % r1.water_radius
    1.50
    >>> r2 = Radii()
    >>> print '%.2f' % r2.water_radius
    1.50
    >>> r1.radii_set = 'LI_AND_NUSSINOV'
    >>> print '%.2f' % r1.water_radius
    1.37
    >>> print '%.2f' % r2.water_radius
    1.37
    >>> r2.radii_set = 'WILLIAMS'
    >>> print '%.2f' % r1.water_radius
    1.50
    >>> print '%.2f' % r2.water_radius
    1.50


    """

    def __init__(self):
        """
        Initialise an instance.
        """
        # Polar Atom
        #       'charApolRadius'   : 1.97,
        #       'charApolRadiusMin': 1.35,
        #       'charApolRadiusMax': 2.59,
        #       'charPolRadius'    : 1.50,
        #       'charPolRadiusMin' : 1.32,
        #       'charPolRadiusMax' : 1.63,
        #       'closeRadius'      : charPolRadius    + self._WILLIAMS_POLAR_CLOSE_CONST,
        #       'farRadius'        : charPolRadiusMax + self._WILLIAMS_POLAR_FAR_CONST,
        #       'probeRadius'      : charApolRadiusMax,
        #
        # Apolar Atom
        #       'charApolRadius'   : 1.93,
        #       'charApolRadiusMin': 1.63,
        #       'charApolRadiusMax': 2.23,
        #       'closeRadius'      : charApolRadiusMin + self._WILLIAMS_APOLAR_CLOSE_CONST,
        #       'farRadius'        : charApolRadius    + self._WILLIAMS_APOLAR_FAR_CONST,
        #       'probeRadius'      : charApolRadiusMax,

        # Water Radii

        self._WILLIAMS_POLAR_CLOSE_CONST  = 1.50  # H2O characteristic polar radius
        self._WILLIAMS_POLAR_FAR_CONST    = 1.63  # H2O maximum polar radius
        self._WILLIAMS_APOLAR_CLOSE_CONST = 1.35  # H2O minimum apolar radius
        self._WILLIAMS_APOLAR_FAR_CONST   = 1.97  # H2O characteristic apolar radius

        # Original radii used by PRO_ACT
        self._WILLIAMS = {
            'H2O': {
                'charApolRadius'   : 1.97,
              # 'charApolRadiusMin': 1.35,
                'charApolRadiusMax': 2.59,
                'charPolRadius'    : 1.50,
              # 'charPolRadiusMin' : 1.32,
                'charPolRadiusMax' : 1.63,
                'closeRadius'      : 1.50 + self._WILLIAMS_POLAR_CLOSE_CONST,
                'farRadius'        : 1.63 + self._WILLIAMS_POLAR_FAR_CONST,
                'probeRadius'      : 2.59,
            },
            'O': {
                'charApolRadius'   : 1.62,
              # 'charApolRadiusMin': 1.18,
                'charApolRadiusMax': 2.06,
                'charPolRadius'    : 1.35,
              # 'charPolRadiusMin' : 1.08,
                'charPolRadiusMax' : 1.62,
                'closeRadius'      : 1.35 + self._WILLIAMS_POLAR_CLOSE_CONST,
                'farRadius'        : 1.62 + self._WILLIAMS_POLAR_FAR_CONST,
                'probeRadius'      : 2.06,
            },
            'N': {
                'charApolRadius'   : 1.82,
              # 'charApolRadiusMin': 1.23,
                'charApolRadiusMax': 2.41,
                'charPolRadius'    : 1.45,
              # 'charPolRadiusMin' : 1.23,
                'charPolRadiusMax' : 1.67,
                'closeRadius'      : 1.45 + self._WILLIAMS_POLAR_CLOSE_CONST,
                'farRadius'        : 1.67 + self._WILLIAMS_POLAR_FAR_CONST,
                'probeRadius'      : 2.41,
            },
            'C': {
                'charApolRadius'   : 1.93,
              # 'charApolRadiusMin': 1.63,
                'charApolRadiusMax': 2.23,
                'closeRadius'      : 1.63 + self._WILLIAMS_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.93 + self._WILLIAMS_APOLAR_FAR_CONST,
                'probeRadius'      : 2.23,
            },
            'S': {
                'charApolRadius'   : 1.68,
              # 'charApolRadiusMin': 1.42,
                'charApolRadiusMax': 1.94,
                'closeRadius'      : 1.42 + self._WILLIAMS_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.68 + self._WILLIAMS_APOLAR_FAR_CONST,
                'probeRadius'      : 1.94,
            },
        # Halogens from:
        # Rowland and Taylor J. Phys. Chem. 1996 100, 7384
        # 'Intermolecular Nonbonded Contact Distances in Organic Crystal Structures:
        # Comparison with Distances Expected from van der Waals Radii'
	# Polar values are rather arbitrarily -0.2 and +0.0 c.f. charApolar
            'F': {
                'charApolRadius'   : 1.46,  # Actual number from ref
              # 'charApolRadiusMin': 1.26,  # -0.20
                'charApolRadiusMax': 1.66,  # +0.20
                'charPolRadius'    : 1.26,
                'charPolRadiusMax' : 1.46,
                'closeRadius'      : 1.26 + self._WILLIAMS_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.46 + self._WILLIAMS_APOLAR_FAR_CONST,
                'probeRadius'      : 1.66,
            },
            'CL': {
                'charApolRadius'   : 1.76,  # Actual number from ref
              # 'charApolRadiusMin': 1.56,  # -0.20
                'charApolRadiusMax': 1.96,  # +0.20
                'charPolRadius'    : 1.56,
                'charPolRadiusMax' : 1.76,
                'closeRadius'      : 1.56 + self._WILLIAMS_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.76 + self._WILLIAMS_APOLAR_FAR_CONST,
                'probeRadius'      : 1.96,
            },
            'BR': {
                'charApolRadius'   : 1.87,  # Actual number from ref
              # 'charApolRadiusMin': 1.67,  # -0.20
                'charApolRadiusMax': 2.07,  # +0.20
                'charPolRadius'    : 1.67,
                'charPolRadiusMax' : 1.87,
                'closeRadius'      : 1.67 + self._WILLIAMS_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.87 + self._WILLIAMS_APOLAR_FAR_CONST,
                'probeRadius'      : 2.07,
            },
            'I': {
                'charApolRadius'   : 2.03,  # Actual number from ref
              # 'charApolRadiusMin': 1.83,  # -0.20
                'charApolRadiusMax': 2.23,  # +0.20
                'charPolRadius'    : 1.83,
                'charPolRadiusMax' : 2.03,
                'closeRadius'      : 1.83 + self._WILLIAMS_APOLAR_CLOSE_CONST,
                'farRadius'        : 2.03 + self._WILLIAMS_APOLAR_FAR_CONST,
                'probeRadius'      : 2.23,
            },
        # From Bondi 'van der Waals Volumes and Radii' J. Phys. Chem 1964, 68, 441-451
            'P': {
                'charApolRadius'   : 1.80,   # Actual number from ref
              # 'charApolRadiusMin': 1.60,   # -0.20
                'charApolRadiusMax': 2.00,   # +0.20
                'closeRadius'      : 1.60 + self._WILLIAMS_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.80 + self._WILLIAMS_APOLAR_FAR_CONST,
                'probeRadius'      : 2.00,
            },
        }

        # Li & Nussinov radii
        # Water Radii
        # Atom r(vdw) r(vdwMin) r(vdwDiff) r(charApolMin) r(charApolMax)
        # O    1.51   1.25      0.26       1.25           1.77    
        # N    1.66   1.33      0.33       1.33           1.99
        # S    1.92   1.60      0.32       1.60           2.24
        # C    1.92   1.70      0.22       1.70           2.14
        # HOH  1.68   1.36      0.32       1.36           2.00
        #
        # Atom r(coul) r(coulMin) r(coulDiff) r(charPolMin) r(charPolMax)
        # O    1.38   1.26        0.12        1.26          1.50
        # N    1.47   1.29        0.18        1.29          1.65
        # HOH  1.37   1.19        0.18        1.19          1.55

        self._LI_AND_NUSSINOV_POLAR_CLOSE_CONST  = 1.37  # H2O characteristic polar radius
        self._LI_AND_NUSSINOV_POLAR_FAR_CONST    = 1.55  # H2O maximum polar radius
        self._LI_AND_NUSSINOV_APOLAR_CLOSE_CONST = 1.36  # H2O minimum apolar radius
        self._LI_AND_NUSSINOV_APOLAR_FAR_CONST   = 1.68  # H2O characteristic apolar radius

        # Original radii used by self._LI_AND_NUSSINOV
        self._LI_AND_NUSSINOV = {
            'H2O': {
                'charApolRadius'   : 1.68,
                'charApolRadiusMax': 2.00,
                'charPolRadius'    : 1.37,
                'charPolRadiusMax' : 1.55,
                'closeRadius'      : 1.37 + self._LI_AND_NUSSINOV_POLAR_CLOSE_CONST,
                'farRadius'        : 1.55 + self._LI_AND_NUSSINOV_POLAR_FAR_CONST,
                'probeRadius'      : 2.00,
            },
            'O': {
                'charApolRadius'   : 1.51,
                'charApolRadiusMax': 1.77,
                'charPolRadius'    : 1.38,
                'charPolRadiusMax' : 1.50,
                'closeRadius'      : 1.38 + self._LI_AND_NUSSINOV_POLAR_CLOSE_CONST,
                'farRadius'        : 1.50 + self._LI_AND_NUSSINOV_POLAR_FAR_CONST,
                'probeRadius'      : 1.77,
            },
            'N': {
                'charApolRadius'   : 1.66,
                'charApolRadiusMax': 1.99,
                'charPolRadius'    : 1.47,
                'charPolRadiusMax' : 1.65,
                'closeRadius'      : 1.47 + self._LI_AND_NUSSINOV_POLAR_CLOSE_CONST,
                'farRadius'        : 1.65 + self._LI_AND_NUSSINOV_POLAR_FAR_CONST,
                'probeRadius'      : 1.99,
            },
            'C': {
                'charApolRadius'   : 1.92,
                'charApolRadiusMax': 2.14,
                'charApolRadiusMin': 1.70,
                'closeRadius'      : 1.70 + self._LI_AND_NUSSINOV_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.92 + self._LI_AND_NUSSINOV_APOLAR_FAR_CONST,
                'probeRadius'      : 2.14,
            },
            'S': {
                'charApolRadius'   : 1.92,
                'charApolRadiusMax': 2.24,
                'charApolRadiusMin': 1.60,
                'closeRadius'      : 1.60 + self._LI_AND_NUSSINOV_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.92 + self._LI_AND_NUSSINOV_APOLAR_FAR_CONST,
                'probeRadius'      : 2.24,
            },
        # Halogens from:
        # Rowland and Taylor J. Phys. Chem. 1996 100, 7384
        # 'Intermolecular Nonbonded Contact Distances in Organic Crystal Structures:
        # Comparison with Distances Expected from van der Waals Radii'
            'F': {
                'charApolRadius'   : 1.46,  # Actual number from ref
              # 'charApolRadiusMin': 1.26,  # -0.20
                'charApolRadiusMax': 1.66,  # +0.20
                'charPolRadius'    : 1.26,
                'charPolRadiusMax' : 1.46,
                'closeRadius'      : 1.26 + self._LI_AND_NUSSINOV_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.46 + self._LI_AND_NUSSINOV_APOLAR_FAR_CONST,
                'probeRadius'      : 1.66,
            },
            'CL': {
                'charApolRadius'   : 1.76,  # Actual number from ref
              # 'charApolRadiusMin': 1.56,  # -0.20
                'charApolRadiusMax': 1.96,  # +0.20
                'charPolRadius'    : 1.56,
                'charPolRadiusMax' : 1.76,
                'closeRadius'      : 1.56 + self._LI_AND_NUSSINOV_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.76 + self._LI_AND_NUSSINOV_APOLAR_FAR_CONST,
                'probeRadius'      : 1.96,
            },
            'BR': {
                'charApolRadius'   : 1.87,  # Actual number from ref
              # 'charApolRadiusMin': 1.67,  # -0.20
                'charApolRadiusMax': 2.07,  # +0.20
                'charPolRadius'    : 1.67,
                'charPolRadiusMax' : 1.87,
                'closeRadius'      : 1.67 + self._LI_AND_NUSSINOV_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.87 + self._LI_AND_NUSSINOV_APOLAR_FAR_CONST,
                'probeRadius'      : 2.07,
            },
            'I': {
                'charApolRadius'   : 2.03,  # Actual number from ref
              # 'charApolRadiusMin': 1.83,  # -0.20
                'charApolRadiusMax': 2.23,  # +0.20
                'charPolRadius'    : 1.83,
                'charPolRadiusMax' : 2.03,
                'closeRadius'      : 1.83 + self._LI_AND_NUSSINOV_APOLAR_CLOSE_CONST,
                'farRadius'        : 2.03 + self._LI_AND_NUSSINOV_APOLAR_FAR_CONST,
                'probeRadius'      : 2.23,
            },
        # From Bondi 'van der Waals Volumes and Radii' J. Phys. Chem 1964, 68, 441-451
            'P': {
                'charApolRadius'   : 1.80,   # Actual number from ref
              # 'charApolRadiusMin': 1.60,   # -0.20
                'charApolRadiusMax': 2.00,   # +0.20
                'closeRadius'      : 1.60 + self._LI_AND_NUSSINOV_APOLAR_CLOSE_CONST,
                'farRadius'        : 1.80 + self._LI_AND_NUSSINOV_APOLAR_FAR_CONST,
                'probeRadius'      : 2.00,
            },
        }

    _radii_set = None

    def get_radii(self):
        """
        Get the radii set state.
        """
        return Radii._radii_set

    def set_radii(self, radii_set):
        """
        Set the radii set state.
        """
        Radii._radii_set = radii_set
        if radii_set == 'WILLIAMS':
            Radii._RADII = self._WILLIAMS
        elif radii_set == 'LI_AND_NUSSINOV':
            Radii._RADII = self._LI_AND_NUSSINOV
        else:
            raise ProAct2Error('Unkown radii set')
        # No need to change the lines below
        Radii._WATER_RADIUS = Radii._RADII['H2O']['charPolRadius']
        Radii._WATER_RADIUS_MAX = Radii._RADII['H2O']['charPolRadiusMax']
        Radii._WATER_APOLAR_RADIUS = Radii._RADII['H2O']['charApolRadiusMax']

    radii_set = property(get_radii, set_radii)
    
    @property
    def radii(self):
        """
        Return the radii dictionary.

        """
        return Radii._RADII

    @property
    def water_radius(self):
        """
        Return the water radius.

        """
        return Radii._WATER_RADIUS

    @property
    def water_radius_max(self):
        """
        Return the maximum water radius.

        """
        return Radii._WATER_RADIUS_MAX

    @property
    def water_apolar_radius(self):
        """
        Return the water apolar radius.

        """
        return Radii._WATER_APOLAR_RADIUS

# Test the class by running the command 'python Radii.py' in the terminal
if __name__ == "__main__":
    import doctest
    doctest.testmod()

