from Point import Point

class SurfacePoint(Point):
    """Class SurfacePoint

    The class SurfacePoint inherits all the attributes and methods of the
    Point class. Furthermore, a surface point has additional attributes
    relating to its distance from the atom centre and its accessibility.

    Examples:
    >>> from SurfacePoint import SurfacePoint
    >>> from Point import Point
    >>> u = Point(0.0, 1.0, 0.0)   # A unit surface point
    >>> c = Point(1.0, 1.0, 1.0)   # The centre (of an atom)
    >>> sp = SurfacePoint(c, u, 0, radius=3.5, accessibility=0, flag=0 )
    >>> print sp
    <1.00, 4.50, 1.00>
    >>> sp.update_surface_point(radius=2.5, x=sp.x, y=sp.y, z=sp.z, accessibility=1, flag=2)
    >>> print sp
    <1.00, 3.50, 1.00>
    >>> sp.accessibility
    1
    >>> sp.flag
    2

    """
    # Slots saves memory if there are millions of instances of an object. The
    # line below can be commented out to get more flexibility out of this
    # class.
    __slots__ = '_centre_point', '_unit_surface_point', \
                '_unit_surface_point_index', '_radius', \
                '_accessibility', '_flag'
    
    # Operations
    def __init__(self, centre_point, unit_surface_point,
                 unit_surface_point_index, radius, accessibility, flag):
        self._centre_point = centre_point
        self._unit_surface_point = unit_surface_point
        self._unit_surface_point_index = unit_surface_point_index
        self._radius = radius
        self._accessibility = accessibility
        self._flag = flag
        x = self._unit_surface_point._x*radius + self._centre_point._x
        y = self._unit_surface_point._y*radius + self._centre_point._y
        z = self._unit_surface_point._z*radius + self._centre_point._z
        Point.__init__(self, x, y, z)

    def get_radius(self):
        """Get radius attribute."""
        return self._radius
    
    def set_radius(self, radius):
        """Set radius attribute."""
        self._radius = float(radius)
        x = self._unit_surface_point._x*radius + self._centre_point._x
        y = self._unit_surface_point._y*radius + self._centre_point._y
        z = self._unit_surface_point._z*radius + self._centre_point._z
        self.x = x
        self.y = y
        self.z = z
    
    radius = property(get_radius, set_radius)

    def get_accessibility(self):
        """Get accessibility attribute."""
        return self._accessibility
    
    def set_accessibility(self, accessibility):
        """Set accessibility attribute."""
        self._accessibility = int(accessibility)
    
    accessibility = property(get_accessibility, set_accessibility)
    
    def get_flag(self):
        """Get flag attribute."""
        return self._flag
    
    def set_flag(self, flag):
        """Set flag attribute."""
        self._flag = int(flag)
    
    flag = property(get_flag, set_flag)
    
    @property
    def unit_surface_point(self):
        """Get unit_surface_point attribute."""
        return self._unit_surface_point
    
    @property
    def unit_surface_point_index(self):
        """Get unit_surface_point_index attribute."""
        return self._unit_surface_point_index

    @property
    def centre_point(self):
        """Get centre_point attribute."""
        return self._centre_point
    
    def update_surface_point(self, radius, x, y, z, accessibility, flag):
        """
        Change the radius accessibility and flag of a surface point.
        
        radius: float
        x, y, z: float
        accessibility: int
        flag: int

        This method sets the following attributes:
        1) _radius
        2) _accessibility
        3) _flag
        4) _x, _y, _z

        """
        self.set_radius(radius)
        self._centre_point.x = x
        self._centre_point.y = y
        self._centre_point.z = z
        self.set_accessibility(accessibility)
        self.set_flag(flag)
    
# test SurfacePoint by the command "python SurfacePoint.py"
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    import unittest
    suite = doctest.DocFileSuite('doc_tests/test_SurfacePoint.txt',
        optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)
    unittest.TextTestRunner().run(suite)
