from Atom import Atom
# Deal with the radii
import Radii
R = Radii.Radii()

class WaterProbe(Atom):
    """Class WaterProbe

    The class WaterProbe inherits all its attributes and methods from the
    class Atom. Furthermore it has got the extra attributes __waterType and
    __hydrationLayer. The water types are defined as:
    1) buried  - (0)
    2) cleft   - (>1)
    3) surface - (1)
    4) further - (<0)
    The 'further' type can only arise when more than one hydration layer is
    added using the Assembly.addWater() method. The negative number refers to
    the distance from the surface layer, in other words a water with a
    __waterType of -1 is one layer ouside of the surface layer.

    The __hydrationLayer attribute refers to when the WaterProbe was added to
    the Assembly. The hydration layers are defined as:
    1) experimental                     - (0)
    2) Assembly.addWater() (first run)  - (1)
    3) Assembly.addWater() (second run) - (2)
    4) etc...

    Examples:
    >>> from WaterProbe import WaterProbe
    >>> import Radii
    >>> R = Radii.Radii()
    >>> R.radii_set = 'LI_AND_NUSSINOV'
    >>> la = WaterProbe(1, '  O ', 'O', 0.0, 0.0, 0.0, 3, 0, 2, True)
    >>> la.water_type = 1
    >>> la.water_type
    1
    >>> la.hydration_layer = 1
    >>> la.hydration_layer
    1
    >>> la.res_original_id
    2

    """
    
    def __init__(self, atom_org_id, atom_org_name, atom_type, x, y, z,
                 sub_atom_type, molecule_instance, res_original_id,
                 experimental):
        """
        Initialise the object.

        atom_org_id: int
        atom_org_name: string
        atom_type: string
        x: float
        y: float
        z: float
        sub_atom_type: int
        molecule_instance: int
        res_original_id: int
        experimental: boolean

        This method sets the following attributes of the atom:
        1) _atom_org_id
        2) _atom_org_name
        3) _atom_type
        4) _x, _y, _z
        5) _sub_atom_type
        6) _molecule_instance
        7) _atom_polarity
        8) various radii...

        """

        Atom.__init__(self, atom_org_id, atom_org_name, atom_type, x, y, z,
                      sub_atom_type, molecule_instance)
        self._set_atom_polarity()
        self._set_radius()
        self._res_original_id = res_original_id
        self._experimental = experimental
        
    @property
    def res_original_id(self):
        """Get res_original_id attribute."""
        return self._res_original_id
    
    def get_water_type(self):
        """Get water_type attribute."""
        return self._water_type
    
    def set_water_type(self, water_type):
        """Set water_type attribute."""
        self._water_type = int(water_type)
    
    water_type = property(get_water_type, set_water_type)
    
    def get_hydration_layer(self):
        """Get hydration_layer attribute."""
        return self._hydration_layer
    
    def set_hydration_layer(self, hydration_layer):
        """Set hydration_layer attribute."""
        self._hydration_layer = int(hydration_layer)
    
    hydration_layer = property(get_hydration_layer, set_hydration_layer)
    
    def _set_radius(self):
        """function _set_radius
        
        This method sets the various atomic radii defined in the Atom class.
        
        """
        radii = R.radii[self.atom_type]           # should always be H20
        self.char_apol_radius = radii['charApolRadius']
        if self.atom_polarity == 0:               # should always be true
            self.char_pol_radius = radii['charPolRadius']
            self.char_pol_radius_max = radii['charPolRadiusMax']
        self.close_radius = radii['closeRadius']
        self.far_radius = radii['farRadius']
        self.probe_radius = radii['probeRadius']
    
    def _set_atom_polarity(self):
        """function _set_atom_polarity
        """
        self.atom_polarity = 0  # waters are polar
    
    @property
    def experimental(self):
        """Get experimental attribute.
        Is it a crystallographic water or not?"""
        return self._experimental
        
# Test WaterProbe class by 'python WaterProbe.py'
if __name__ == '__main__':
    import doctest
    doctest.testmod()
    import unittest
    suite = doctest.DocFileSuite('doc_tests/test_WaterProbe.txt',
        optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)
    unittest.TextTestRunner().run(suite)
