#!/usr/bin/env python

import sys
import os.path

# The psyco import could be commented out if it does not have any significant benefit.
try:
    import psyco
    psyco.full()
except ImportError:
    sys.stderr.write('Psyco not installed. proACT2 will run without acceleration\n')

import optparse
import random
import re
import copy
import time
import mol2
import Ligand
from Bio.PDB.PDBParser import PDBParser
import MacroMolecule
import Hydration
import ContactList
import Assembly
import UnitSurfacePoints
from ProAct2Error import ProAct2Error
import Radii
R = Radii.Radii()


class FileFaker:
    """
    This is a hack to overcome the fact that the pdb parser
    writes it's errors to stdout
    """
    def write(self, string):
        sys.stderr.write(string)

class ProAct(object):
    """
    This class controls the flow of the pro_act program.
    """
    
    def __init__(self):
        """
        Initialise a ProAct instance.
        """
        self.A = Assembly.Assembly()
        self.U = UnitSurfacePoints.UnitSurfacePoints()
        self.PI = 3.14159

    def _chain_sel_to_list(self, x):
        if x == 'all':
            return 'all'
        x = x.strip()  # remove preceeding and trailing whitespaces
        x = x.upper()  # upper case
        list = []      # make list
        for letter in x:
            list.append(letter)
        return list

    def _res_sel_to_list(self, x):
        if x == 'all':
            return 'all'
        x = x.strip()  # remove preceeding and trailing whitespaces
        xlist = x.split(',')  # split selections
        list = []
        for selection in xlist:
            if selection.find(':') == -1:  # not a range
                res_id = int(selection)     # convert to int
                list.append(res_id)                 # append to list
            else:                                       # a range
                indices = selection.split(':')  # get the indices
                start = int(indices[0])         # convert to int
                stop = int(indices[1])          # convert to int
                stop = stop + 1                         # be inclusive
                for i in range(start, stop):
                    list.append(i)
        return list

    def _read_command_line(self):
        """
        Get command line options and arguements.

        """

        usage = "%prog [options] file1 [file2]"
        version = "%prog 2.0"
        parser = optparse.OptionParser(usage=usage, version=version)
        parser.add_option('--runs',
                          dest='n_runs',
                          default='1',
                          help='number of runs (default 1)')
        parser.add_option('--radii',
                          dest='radii_set',
                          default='LI_AND_NUSSINOV',
                          help='radii set used; WILLIAMS or LI_AND_NUSSINOV (default LI_AND_NUSSINOV)')
        parser.add_option('--cavity_min_radius',
                          dest='probe_min_cutoff',
                          default=0.8,
                          help='minimum radius of any cavities to be detected (default 0.8)')
        parser.add_option('--water_placement',
                          dest='water_placement',
                          default='proact2',
                          help='water placement algorithm; proact2, proact1 or random (default proact2)')
        parser.add_option('--experimental_waters',
                          dest='experimental_waters',
                          default=False,
                          action="store_true",
                          help='include experimental waters (default False)')
        parser.add_option('--aggressive_waters',
                          dest='aggressive_waters',
                          default=False,
                          action="store_true",
                          help='add waters agressively in cavities')
        parser.add_option('--layers',
                          dest='n_hydration',
                          default='1',
                          help='number of hydration layers (default 1)')
        parser.add_option('--only_binding_site',
                          dest='only_bs',
                          default=False,
                          action='store_true',
                          help='only use binding site atoms')
        parser.add_option('--interface_type',
                          dest='bs_type',
                          default='proact2',
                          help='binding site/interface algorithm; proact2 or cutoff e.g. --interface_type=5.6 (default proact2)')
        parser.add_option('--only_interface',
                          dest='only_interface',
                          default=False,
                          action='store_true',
                          help='only use protein-protein interface atoms')
        parser.add_option('--complex_formation',
                          dest='return_interaction',
                          default=False,
                          action='store_true',
                          help='return changes occuring on formation of a complex between two molecules (P+L)-P.L')
        parser.add_option('--res_sel1',
                          dest='res_sel1',
                          default='all',
                          help='residue selection for first file; e.g. 1,3,6:9')
        parser.add_option('--res_sel2',
                          dest='res_sel2',
                          default='all',
                          help='residue selection for second file; e.g. 1,3,6:9')
        parser.add_option('--chain_sel1',
                          dest='chain_sel1',
                          default='all',
                          help='chain selection for first structure file; e.g. A or ABC')
        parser.add_option('--chain_sel2',
                          dest='chain_sel2',
                          default='all',
                          help='chain selection for second structure file; e.g. A or ABC')
        parser.add_option('--file_type1',
                          dest='file_type1',
                          help='explicitly set the file type of first file (pdb or mol2)')
        parser.add_option('--file_type2',
                          dest='file_type2',
                          help='explicitly set the file type of second file (pdb or mol2)')
        parser.add_option('--kinemage',
                          dest='kin_output',
                          default=False,
                          action='store_true',
                          help='Write out kinemage files.')
        parser.add_option('--surface',
                          dest='surface',
                          default=False,
                          action='store_true',
                          help='Write out dms surface file.')
        parser.add_option('--no_naccess',
                          dest='naccess_output',
                          default=True,
                          action='store_false',
                          help='Do not write out naccess files.')
        parser.add_option('--no_summary',
                          dest='summary_output',
                          default=True,
                          action='store_false',
                          help='Do not write out summary file.')
        parser.add_option('--no_water_pdb',
                          dest='water_output',
                          default=True,
                          action='store_false',
                          help='Do not write out water pdb file.')
        parser.add_option('--no_cavity_pdb',
                          dest='cavity_output',
                          default=True,
                          action='store_false',
                          help='Do not write out cavity pdb file.')
        parser.add_option('--no_residue_residue_csv',
                          dest='res_res_output',
                          default=True,
                          action='store_false',
                          help='Do not write residue-residue contact csv file.')
       
        options, args = parser.parse_args(sys.argv[1:])

        if len(args) < 1 or len(args) > 2:
            parser.error('Incorrect number of arguments')

        known_file_types = ['pdb', 'mol2']
        if not options.file_type1:
            options.file_type1 = args[0].split('.')[-1].strip()
            if options.file_type1 == 'ent':
                options.file_type1 = 'pdb'
        if not options.file_type2 and len(args) == 2:
            options.file_type2 = args[1].split('.')[-1].strip()
            if options.file_type2 == 'ent':
                options.file_type1 = 'pdb'
        if options.file_type1 not in known_file_types:
            parser.error('Unknown file type: %s' % options.file_type1)
        if options.file_type2 not in known_file_types and len(args) == 2:
            parser.error('Unknown file type: %s' % options.file_type2)

        if len(args) != 2:
            if options.only_bs:
                parser.error('Option only_binding_site requires two structure files')
            if options.only_interface:
                parser.error('Option only_interface requires two structure files')
            if options.return_interaction:
                parser.error('Option complex_formation requires two structure files')
        
        if options.only_bs and options.only_interface:
            parser.error('The options only_binding_site and only_interface are mutually exclusive')

        if options.only_bs:
            if options.file_type1 != 'pdb':
                parser.error("option only_binding_site requires first file to be 'pdb'")
            if options.file_type2 != 'mol2':
                parser.error("option only_binding_site requires second file to be 'mol2'")
            
        if options.only_interface:
            if options.file_type1 != 'pdb':
                parser.error("option only_interface requires first file to be 'pdb'")
            if options.file_type2 != 'pdb':
                parser.error("option only_interface requires second file to be 'pdb'")
        
        if options.water_placement not in ['proact2', 'proact1', 'random']:
            parser.error("unrecognized argument for option water_placement")

        if options.radii_set not in ['LI_AND_NUSSINOV', 'WILLIAMS']:
            parser.error("unrecognized argument for option radii_set")
        R.radii_set = options.radii_set

        if options.experimental_waters and options.naccess_output:
            options.naccess_output = False
            sys.stderr.write('Warning: cannot use naccess output option when including experimental waters, naccess output switched off\n')

        if options.experimental_waters and options.surface:
            options.surface = False
            sys.stderr.write('Warning: cannot use surface output option when including experimental waters, surface output switched off\n')

        options.chain_sel1 = self._chain_sel_to_list(options.chain_sel1)
        options.chain_sel2 = self._chain_sel_to_list(options.chain_sel2)

        options.res_sel1 = self._res_sel_to_list(options.res_sel1)
        options.res_sel2 = self._res_sel_to_list(options.res_sel2)

        options.n_runs = int(options.n_runs)
        options.n_hydration = int(options.n_hydration)
        options.probe_min_cutoff = float(options.probe_min_cutoff)

        if len(args) == 1:
            args.append(None)
            options.file_type2 = None
            options.chain_sel2 = None
            options.res_sel2 = None

        info = """
        ------------------------------------------------------
                   Structure File 1
        File                       : %s
        Type                       : %s
        Chain Selection            : %s
        Residue Selection          : %s
        ------------------------------------------------------
                   Structure File 2
        File                       : %s
        Type                       : %s
        Chain Selection            : %s
        Residue Selection          : %s
        ------------------------------------------------------
                            Options
        Radii Set                  : %s
        Cavity min radius          : %.2f
        Only Binding Site          : %s
        Binding Site Definition    : %s
        Only Interface             : %s
        Complex Formation          : %s
        Number Of Runs             : %i
        Number Of Hydration Layers : %i
        Water Placement Method     : %s
        Aggressive Waters          : %s
        Experimental Waters        : %s
        ------------------------------------------------------
            """ % (args[0],
                   options.file_type1,
                   options.chain_sel1,
                   options.res_sel1,
                   args[1],
                   options.file_type2,
                   options.chain_sel2,
                   options.res_sel2,
                   options.radii_set,
                   options.probe_min_cutoff,
                   options.only_bs,
                   options.bs_type,
                   options.only_interface,
                   options.return_interaction,
                   options.n_runs,
                   options.n_hydration,
                   options.water_placement,
                   options.aggressive_waters,
                   options.experimental_waters)
        
        self.args = args
        self.options = options
        self.info = re.sub(r'\n\s+', '\n',  info)

    def _read_mol2(self, molecule_instance, file_name):
        """function read_mol2
        
        molecule_instance: int
        file_name: string

        This method reads the mol2 file and returns a Ligand.
        
        returns Ligand
        """
        L = Ligand.Ligand()
        ligand = mol2.Mol2(file_name)
        for atom in ligand.atom_rec:
            if atom['atom_type'] == 'H':
                continue
            L.atom_input(
                atom['atom_id'],
                atom['atom_name'],
                atom['atom_name'],
                atom['x'],
                atom['y'],
                atom['z'],
                2,
                molecule_instance)
        return L

    def _read_pdb(self, molecule_instance, file_name, residue_selection, chain_selection):
        """
        Read in a PDB file.
        
        molecule_instance: int
        file_name: string
        residue_selection: list
        chain_selection: list
        includeExpWaters: boolean

        This method reads in all the atoms of the PDB file specified by the
        residue_selection and the chain_selection. They should have the
        following form:
        
        1) chain_selection= 'A' (chain A) or 'AB' (chain A and chain B)
        
        2) residue_selection= '3' (residue 3) or '3-5' (residues 3,4 and 5)
                             or '3-5,10' (residues 3,4,5 and 10)

        If the includeExpWaters is set to True the method will read the
        experimental waters (of the selected chains/residues) into the
        Hydration class.

        In cases of disorder the Bio.PDB module automatically selects the atom
        with the highest occupancy.

        Protein atoms will be read into the MacroMolecule class.

        All HETAtoms except water are currently excluded.

        Nucleic acids are currently not supported.
        
        returns MacroMolecule

        """

        M = MacroMolecule.MacroMolecule()
        H = Hydration.Hydration()
        
        # This is a hack to overcome the fact that the pdb parser
        # writes it's errors to stdout
        sys.stdout = FileFaker()
        
        p = PDBParser(PERMISSIVE=1)  # load the pdb parser object
        structure_id = file_name.replace('.pdb', '')
        structure = p.get_structure(structure_id, file_name)

        # Remove the hack
        sys.stdout = sys.__stdout__

        model = structure[0]

        atom_id = 0
        for chain in model.get_list():
            if chain_selection == 'all':
                use_chain = True
            else:
                use_chain = False
                for c in chain_selection:
                    if chain.get_id() == c:
                        use_chain = True
                        break
            if not(use_chain):  # do not use this chain
                continue       # next chain
            for residue in chain.get_list():
                if residue_selection == 'all':
                    use_res = True
                else:
                    use_res = False
                    for r in residue_selection:
                        if residue.get_id()[1] == r:
                            use_res = True
                            break
                if residue.get_id()[0] == 'W':  # don't include waters
                    use_res = False
                    if self.options.experimental_waters:  # unless told to
                        use_res = True
                if residue.get_id()[0].find('H_') != -1:  # hetatm
                    use_res = False                        # exclude
                if not(use_res):  # do not use this residue
                    continue     # next residue
                for atom in residue.get_list():

                    # Get atom type
                    atom_type = atom.get_fullname()[1]

                    # Mild sanity check of atom names
                    if atom_type == 'H':   # hydrogen exclude
                        continue
                    if atom_type == 'D':   # deuterium exclude
                        continue
                    if re.match(r'H', atom.get_fullname()) != None:
                        continue          # exclude weird hydrogens
                    if re.match(r'^D', atom.get_fullname()) != None:
                        continue          # exclude deuterium 
                        
                    if ( atom_type != 'C'
                        and atom_type != 'N'
                        and atom_type != 'O'
                        and atom_type != 'S'
                        and atom_type != 'H'):
                        a_type = atom.get_fullname()
                        errorMessage = "Atom type?: %s %i" (a_type,
                                                        residue.get_id()[0])
                        raise ProAct2Error(errorMessage)

                    if residue.get_id()[0] == 'W':
                        atom_id += 1
                        H.atom_input(
                            1,
                            residue.get_id()[1],
                            atom_id,               # can't get the real atom id
                            '  O ',                # atomName
                            'H2O',
                            atom.get_coord()[0],   # x
                            atom.get_coord()[1],   # y
                            atom.get_coord()[2],   # z
                            3,                     # subAtomType
                            molecule_instance,
                            True)
                        H[-1].include = 1
                    else:
                        atom_id += 1
                        M.atom_input(
                            residue.get_id()[1],   # residueId
                            residue.get_resname(), # resType
                            chain.get_id(),        # chain
                            atom_id,               # can't get the real atom id
                            atom.get_fullname(),   # atomName
                            atom_type,
                            atom.get_coord()[0],   # x
                            atom.get_coord()[1],   # y
                            atom.get_coord()[2],   # z
                            1,                     # subAtomType
                            molecule_instance)
        return M, H

    def _run(self, system, runs, n_hydration, file_prefix):
        """
        Generate the results.

        system  - the system includes all the atom in a system, for
                  example: for a binding site to correctly set the
                  accessibility of the atoms in the binding site it
                  relys on the other atom in the protein (the system)
        """

        if __debug__:
            log = open('proact2.log', 'a')

        # Open file for saving the runs
        run_file_name = "all_runs.csv"
        if not os.path.isfile(run_file_name):
            run_file = open(run_file_name, "w")
            header = '"structure","run","protein1_atoms","protein2_atoms","ligand_atoms","total_contacts","polar_contacts","apolar_contacts","polar_water_contacts","apolar_water_contacts","polar_surface_water_contacts","apolar_surface_water_contacts","polar_cleft_water_contacts","apolar_cleft_water_contacts","polar_buried_water_contacts","apolar_buried_water_contacts","polar_further_water_contacts","apolar_further_water_contacts","polar_cavity_contacts","apolar_cavity_contacts","surface_waters","cleft_waters","buried_waters","further_waters"\n'
            run_file.write(header)
            run_file.close()
        run_file = open(run_file_name, "a")

        # Initiate contact container
        contacts = ContactList.ContactList()
        contacts.calculate_contact_counts()  # defaults to zero

        # Number of waters
        surface_waters = 0
        cleft_waters = 0
        buried_waters = 0
        further_waters = 0

        # Probe volume
        probe_volume = 0.0
        
        # Run several times in order to get average water contact counts
        for run in range(runs):
            if __debug__:
                start = time.time()
                log.write('Run %i of %i\n' % (run+1, runs))
            run_file.write('"%s",%i,' % (file_prefix, run+1))

            M = self.A.macro_molecule
            L = self.A.ligand
            H = self.A.hydration
            V = self.A.void
            I = self.A.interaction

            # Include experimental waters in the system
            if self.options.experimental_waters:
                system.extend(H)

            tmp_atoms = system[:]
            # environment   - all the atoms which are important to the
            #                 atom_selection
            #                 system = atom_selection + environment
            environment = []
            # atom_selection - the atoms of interest (usually the entire protein)
            #                 for protein-ligand complexes with the binding site
            #                 option set this means: the ligand and the protein
            #                 binding site
            atom_selection = []
            while tmp_atoms:
                atom = random.choice(tmp_atoms)
                if atom.include == 1:
                    atom_selection.append(atom)
                else:
                    environment.append(atom)
                tmp_atoms.remove(atom)
        
            # Set the atoms neighbors
            if __debug__:
                log.write("Setting the neighbors... ")
                tmp_start = time.time()
            Assembly.set_neighbors(system)
            if __debug__:
                tmp_time_taken = time.time() - tmp_start
                log.write('%.2f seconds OK\n' % tmp_time_taken)

            # Set the accessibility of the atoms
            if __debug__:
                log.write("Setting the accessibility... ")
                tmp_start = time.time()
            for atom in atom_selection:
                atom.set_surface_points(self.U, atom.close_radius, 0, 0)
                atom.set_accessibility(self.U)
            if __debug__:
                tmp_time_taken = time.time() - tmp_start
                log.write('%.2f seconds OK\n' % tmp_time_taken)

            # Add the waters
            if __debug__:
                log.write("Adding the water... ")
                tmp_start = time.time()

            next_hoh_id = self.comp_hoh_start_id
            for i in range(n_hydration):
                layer = i + 1
                tmp_environment = environment[:]
                # Included waters should not be used with the ligand
                if file_prefix.find('ligand') == -1:
                    tmp_environment += H
                if self.options.water_placement == 'proact2':
                    next_hoh_id = self.A.add_water(self.U,
                                                   atom_selection,
                                                   tmp_environment,
                                                   layer,
                                                   next_hoh_id,
                                                   self.options.only_interface)
                elif self.options.water_placement == 'proact1':
                    next_hoh_id = self.A.add_water_original(self.U,
                                                            atom_selection,
                                                            tmp_environment,
                                                            layer,
                                                            next_hoh_id,
                                                            self.options.only_interface)
                elif self.options.water_placement == 'random':
                    next_hoh_id = self.A.add_water_random(self.U,
                                                          atom_selection,
                                                          tmp_environment,
                                                          layer,
                                                          next_hoh_id,
                                                          self.options.only_interface)
                else:
                    errorMessage = "Unknown water placement algorithm: %r" % self.options.water_placement
                    raise ProAct2Error(errorMessage)
                    
            water_list = H
            if __debug__:
                tmp_time_taken = time.time() - tmp_start
                log.write('%.2f seconds OK\n' % tmp_time_taken)

            # Determine the water types
            if __debug__:
                log.write("Determining the water types... ")
                tmp_start = time.time()
            self.A.set_water_type(water_list, self.U, self.options.n_hydration)
            if __debug__:
                tmp_time_taken = time.time() - tmp_start
                log.write('%.2f seconds OK\n' % tmp_time_taken)

            # Correct the accessibility maps
            if __debug__:
                log.write("Correcting the accessibility maps... ")
                tmp_start = time.time()
            self.A.correct_water_maps(water_list, atom_selection)
            if __debug__:
                tmp_time_taken = time.time() - tmp_start
                log.write('%.2f seconds OK\n' % tmp_time_taken)

            # Add the cavity pseudo-atoms
            if __debug__:
                log.write("Adding the cavity pseudo-atoms... ")
                tmp_start = time.time()
            self.A.probe_fill(self.U,
                              atom_selection + water_list,
                              self.options.probe_min_cutoff)
            probe_list = V
            if __debug__:
                tmp_time_taken = time.time() - tmp_start
                log.write('%.2f seconds OK\n' % tmp_time_taken)

            if self.options.aggressive_waters:
                # Fill solvatable voids with buried waters
                all_atoms = ( M + L + H + V)
                next_hoh_id = self.A.add_void_waters1(next_hoh_id, probe_list, all_atoms)
                all_atoms = ( M + L + H + V)
                next_hoh_id = self.A.add_void_waters2(next_hoh_id, probe_list, all_atoms)
            
            # Determine the water types
            if __debug__:
                log.write("Determining the water types... ")
                tmp_start = time.time()
            self.A.set_water_type(water_list, self.U, self.options.n_hydration)
            if __debug__:
                tmp_time_taken = time.time() - tmp_start
                log.write('%.2f seconds OK\n' % tmp_time_taken)

            # Set the contacts
            if __debug__:
                log.write("Setting the contacts... ")
                tmp_start = time.time()
            contact_atoms = atom_selection + water_list + probe_list
            self.A.set_contacts(atom_selection, contact_atoms)
            I.calculate_contact_counts()
            if __debug__:
                tmp_time_taken = time.time() - tmp_start
                log.write('%.2f seconds OK\n' % tmp_time_taken)

            # Do something with output
            if __debug__:
                log.write("Printing files... ")

            # Print the output
            prefix = '%s_run%i_' % (file_prefix, run+1)

            if self.options.kin_output:
                acc_file_name = prefix + "accessibility.kin"
                surf_file_name = prefix + "surface.kin"
                wat_file_name = prefix + "waters.kin"
                probe_file_name = prefix + "probes.kin"
                contact_file_name = prefix + "contacts.kin"

                acc_file = open(acc_file_name, mode="w")
                wat_file = open(wat_file_name, mode="w")
                probe_file = open(probe_file_name, mode="w")

            if self.options.naccess_output:
                naccess_pdb_file_name = prefix + "access.pdb"
                naccess_rsa_file_name = prefix + "access.rsa"

                naccess_pdb_file = open(naccess_pdb_file_name, mode="w")
                naccess_rsa_file = open(naccess_rsa_file_name, mode="w")

            if self.options.surface:
                dms_surface_file_name = prefix + "surface.dms"
                dms_surface_file = open(dms_surface_file_name, mode="w")

            if self.options.summary_output:
                sum_file_name = prefix + "summary.txt"
                sum_file = open(sum_file_name, mode="w")

            if self.options.water_output:
                file_name = prefix + "waters.pdb"
                water_pdb_file = open(file_name, mode="w")

            if self.options.cavity_output:
                file_name = prefix + "cavity.pdb"
                cavity_pdb_file = open(file_name, mode="w")

            if self.options.res_res_output:
                file_name = prefix + "residue_contacts.csv"
                res_res_csv_file = open(file_name, mode="w")

            if self.options.kin_output:
                if self.options.only_bs and self.options.return_interaction:
                    contact_file = open(contact_file_name, mode="w")
                    contact_text = self.A.get_kin_binding_site_contacts()
                    contact_file.write(contact_text)
                    contact_file.close()
                elif self.options.only_interface and self.options.return_interaction:
                    contact_file = open(contact_file_name, mode="w")
                    contact_text = self.A.get_kin_interface_contacts()
                    contact_file.write(contact_text)
                    contact_file.close()
                else:
                    contact_file = open(contact_file_name, mode="w")
                    contact_text = self.A.get_kin_contacts()
                    contact_file.write(contact_text)
                    contact_file.close()
                

            if self.options.kin_output:
                acc_text = self.A.get_kin_accessibility(atom_selection)
                acc_file.write(acc_text)
                acc_file.close()

                wat_text = self.A.get_kin_waters(water_list)
                wat_file.write(wat_text)
                wat_file.close()

                probe_text = self.A.get_kin_probes(probe_list)
                probe_file.write(probe_text)
                probe_file.close()

            if self.options.naccess_output:
                pdb_text, rsa_text = self.A.get_naccess_accessibility(self.U, atom_selection)
                
                naccess_pdb_file.write(pdb_text)
                naccess_pdb_file.close()

                naccess_rsa_file.write(rsa_text)
                naccess_rsa_file.close()

            if self.options.surface:
                dms_text = self.A.get_dms_surface(self.U, atom_selection)
                
                dms_surface_file.write(dms_text)
                dms_surface_file.close()

            if self.options.water_output:
                text = self.A.get_pdb_waters(water_list)
                water_pdb_file.write(text)
                water_pdb_file.close()

            if self.options.cavity_output:
                text = self.A.get_pdb_probes(probe_list)
                cavity_pdb_file.write(text)
                cavity_pdb_file.close()

            if self.options.res_res_output:
                text = self.A.get_residue_residue_contacts()
                res_res_csv_file.write(text)
                res_res_csv_file.close()


            if __debug__:
                log.write("OK\n")
            
            # add to contact container
            contacts = I + contacts

            # add to water
            run_buried_waters = 0
            run_surface_waters = 0
            run_cleft_waters = 0
            run_further_waters = 0

            for water in H:
                if water.water_type == 0 or water.water_type == 0.1:
                    buried_waters += 1
                    run_buried_waters += 1
                elif water.water_type == 1:
                    surface_waters += 1
                    run_surface_waters += 1
                elif water.water_type > 1:
                    cleft_waters += 1
                    run_cleft_waters += 1
                else:
                    further_waters += 1
                    run_further_waters += 1

            # add to probe volume
            run_probe_volume = 0.0
            radius_cubed = 0.0
            for probe in V:
                tmp_radius = probe.probe_radius
                radius_cubed = radius_cubed + ( tmp_radius * tmp_radius * tmp_radius )
            run_probe_volume += (4/3) * self.PI * radius_cubed
            probe_volume += (4/3) * self.PI * radius_cubed

            # Write to summary file
            if self.options.summary_output:
                summary_text = []
                summary_text.append(self.A.get_number_of_atoms())
                summary_text.append(I.contacts_string())
                summary_text.append('Probe volume                           : %.3f' % run_probe_volume)
                summary_text.append('Surface waters                         : %i' % run_surface_waters)
                summary_text.append('Cleft waters                           : %i' % run_cleft_waters)
                summary_text.append('Buried waters                          : %i' % run_buried_waters)
                summary_text.append('Further waters                         : %i' % run_further_waters)
                sum_file.write('\n'.join(summary_text))
                sum_file.close()
            
            # Write to run file
            run_file.write(self.A.get_number_of_atoms_csv())
            run_file.write(I.contacts_csv())
            run_file.write("," + str(round(run_probe_volume, 3)))
            run_file.write("," + str(run_surface_waters))
            run_file.write("," + str(run_cleft_waters))
            run_file.write("," + str(run_buried_waters))
            run_file.write("," + str(run_further_waters))
            run_file.write("\n")

            # Delete convenient lists of references
            del tmp_atoms
            del environment
            del atom_selection
            del water_list
            del probe_list
            del contact_atoms
            try:
                del all_atoms
            except UnboundLocalError:
                pass

            # Clear all neighbor references
            for atom in ( M + L):
                atom.clear_neighbors()
                
            # Clear all surfaces
            for atom in ( M + L):
                try:
                    # If the atom is not included in the binding site or the
                    # protein-protein interface it will not have a surface
                    atom.clear_surface()
                except AttributeError:
                    pass
                    
            # Clear the __hydration, __void, and __contact attributes
            H.clear()
            V.clear()

            # Clear the contacts
            I.clear()

            # Run time
            if __debug__:
                time_taken = time.time() - start
                log.write('Run took: %.2f seconds\n' % time_taken)

            # End of run!
            
        # Close run file
        run_file.close()

        contacts = contacts / float(runs)
        surface_waters = surface_waters / float(runs)
        cleft_waters = cleft_waters / float(runs)
        buried_waters = buried_waters / float(runs)
        further_waters = further_waters / float(runs)
        probe_volume = probe_volume / float(runs)

        return (contacts,
            surface_waters,
            cleft_waters,
            buried_waters,
            further_waters,
            probe_volume)

    def main(self):
        """
        This is the function that executes the rest of the program.
        """

        if __debug__:
            log = open('proact2.log', 'w')

        # Read the command line
        self._read_command_line()
        if __debug__:
            log.write(self.info)
        
        # Use first half of file1 as output file prefix
        first_file_name = os.path.split(self.args[0])
        file_prefix = first_file_name[1].split('.')[0]
        
        # Read in the first file
        if self.options.file_type1 == 'pdb':
            M, H = self._read_pdb(molecule_instance=1,
                           file_name=self.args[0],
                           residue_selection=self.options.res_sel1,
                           chain_selection=self.options.chain_sel1)
            self.A.macro_molecule.extend(M)
            self.A.hydration.extend(H)
        elif self.options.file_type1 == 'mol2':
            mol2 = self._read_mol2(molecule_instance=1,
                             file_name=self.args[0])
            self.A.ligand.extend(mol2)
        else:
            raise ProAct2Error('Unknown file type')

        # Read in the second file
        if self.args[1]:
            if self.options.file_type2 == 'pdb':
                M, H = self._read_pdb(molecule_instance=2,
                               file_name=self.args[1],
                               residue_selection=self.options.res_sel2,
                               chain_selection=self.options.chain_sel2)
                self.A.macro_molecule.extend(M)
                self.A.hydration.extend(H)
            elif self.options.file_type2 == 'mol2':
                mol2 = self._read_mol2(molecule_instance=2,
                                 file_name=self.args[1])
                self.A.ligand.extend(mol2)
            else:
                raise ProAct2Error('Unknown file type')

        #  Find out what the number of the first computational water residue
        #  should be
        try:
            self.comp_hoh_start_id = 1 + max([x.res_original_id for x in self.A.hydration])
        except ValueError:
            # No experimental waters
            self.comp_hoh_start_id = 1

        # Set variable for checking whether the only_bs/only_interface
        # option has been set
        only_interface = False

        # If the only_bs option has been ticked set the binding site
        if self.options.only_bs:
            if __debug__:
                start = time.time()
                log.write('Setting the binding site... ')
            if self.options.bs_type == 'proact2':
                self.A.set_binding_site(self.U)
            else:
                try:
                    cutoff = float(self.options.bs_type)
                    self.A.set_binding_site_using_cutoff(self.U, cutoff)
                except ValueError:
                    raise ProAct2Error('Unrecognised cutoff definition')
            only_interface = True
            if __debug__:
                time_taken = time.time() - start
                log.write('%.2f seconds OK\n' % time_taken)
            
        # If the only_interface option has been ticked set the interface
        if self.options.only_interface:
            if __debug__:
                start = time.time()
                log.write('Setting the interface... ')
            self.A.set_interface(self.U)
            only_interface = True
            if __debug__:
                time_taken = time.time() - start
                log.write('%.2f seconds OK\n' % time_taken)
            

        # Are we going to return a simple run of the pro_act algorithm or the
        # results of an interaction between a protein and a ligand or a
        # protein-protein complex?

        # The results of an interaction...
        if self.options.return_interaction:
            if __debug__:
                log.write('Running the main script:\n')
                log.close()

            # This section deals with the ligand (or in the case of a
            # protein-protein interaction the second protein)

            # The interaction can either be a protein-ligand or a
            # protein-protein interaction. The ligand has its atoms A.ligand.
            # In case of a protein-protein interaction the second protein will
            # be loaded in A.protein and the relevant atoms will have a
            # molecule_instance == 2. These are mutually exclusive. Perhaps
            # this code could be more robust.
            atoms = self.A.ligand
            if len(atoms) == 0:
                for atom in self.A.macro_molecule:
                    if atom.molecule_instance == 2:
                        atoms.append(atom)

            (ligand_average_contacts,
             ligand_surface_waters,
             ligand_cleft_waters,
             ligand_buried_waters,
             ligand_further_waters,
             ligand_probe_volume) = self._run(atoms, 
                                 self.options.n_runs,
                                 self.options.n_hydration,
                                 '%s_%s' % (file_prefix, 'ligand'))
            

            # This section deals with the protein (or in the case of a
            # protein-protein in interaction the first protein)
            for atom in self.A.macro_molecule:
                if atom.molecule_instance == 1:
                    atoms.append(atom)

            (protein_average_contacts,
             protein_surface_waters,
             protein_cleft_waters,
             protein_buried_waters,
             protein_further_waters,
             protein_probe_volume) = self._run(atoms, 
                                 self.options.n_runs,
                                 self.options.n_hydration,
                                 '%s_%s' % (file_prefix, 'protein'))

            # This section deals with the protein-ligand complex
            atoms = self.A.macro_molecule + self.A.ligand
            (complex_average_contacts,
             complex_surface_waters,
             complex_cleft_waters,
             complex_buried_waters,
             complex_further_waters,
             complex_probe_volume) = self._run(atoms, 
                                 self.options.n_runs,
                                 self.options.n_hydration,
                                 '%s_%s' % (file_prefix, 'complex'))

            # This section deals with the calculation:
            # protein ligand complex - (ligand + protein)
            interaction = ContactList.ContactList()
            interaction.calculate_contact_counts()
            interaction = interaction + complex_average_contacts
            interaction = interaction - ligand_average_contacts
            interaction = interaction - protein_average_contacts

            # ... and the waters...
            interaction_surface_waters = ( complex_surface_waters
                                         - ligand_surface_waters
                                         - protein_surface_waters )
            interaction_cleft_waters = ( complex_cleft_waters
                                         - ligand_cleft_waters
                                         - protein_cleft_waters )
            interaction_buried_waters = ( complex_buried_waters
                                         - ligand_buried_waters
                                         - protein_buried_waters )
            interaction_further_waters = ( complex_further_waters
                                         - ligand_further_waters
                                         - protein_further_waters )
            
            # ... and the volume...
            interaction_probe_volume = ( complex_probe_volume
                                         - ligand_probe_volume
                                         - protein_probe_volume )


            # Open file for saving the runs
            run_file_name = "all_runs.csv"
            run_file = open(run_file_name, "a")
            run_file.write('"%s.interaction","%s",' % (file_prefix, 'average'))
            sum_file_name = file_prefix + "summary.txt"
            sum_file_name = '%s_interaction_summary.txt' % file_prefix
            sum_file = open(sum_file_name, mode="w")

            # Write to summary file
            summary_text = []
            summary_text.append(self.A.get_number_of_atoms())
            summary_text.append(interaction.contacts_string())
            summary_text.append('Probe volume                           : %.4f' % interaction_probe_volume)
            summary_text.append('Surface waters                         : %i' % interaction_surface_waters)
            summary_text.append('Cleft waters                           : %i' % interaction_cleft_waters)
            summary_text.append('Buried waters                          : %i' % interaction_buried_waters)
            summary_text.append('Further waters                         : %i' % interaction_further_waters)
            sum_file.write('\n'.join(summary_text))
            sum_file.close()
            
            # Write to run file
            run_file.write(self.A.get_number_of_atoms_csv())
            run_file.write(interaction.contacts_csv())
            run_file.write("," + str(round(interaction_probe_volume, 4)))
            run_file.write("," + str(interaction_surface_waters))
            run_file.write("," + str(interaction_cleft_waters))
            run_file.write("," + str(interaction_buried_waters))
            run_file.write("," + str(interaction_further_waters))
            run_file.write("\n")
            run_file.close()

        else:
            if __debug__:
                log.write('Running the main script:\n')
                log.close()
            atoms = self.A.macro_molecule + self.A.ligand
            (average_contacts,
             surface_waters,
             cleft_waters,
             buried_waters,
             further_waters,
             probe_volume) = self._run(atoms, 
                                 self.options.n_runs,
                                 self.options.n_hydration,
                                 file_prefix)

if __name__ == '__main__':
    P = ProAct()
    P.main()

# BUGS
# 1. Biopython does not seem to be able to deal with deuturated waters from
#    neutron structures.
