#!/usr/bin/env python

"""
Author     : Tjelvar Olsson
Email      : tjelvar.olsson@gmail.com
Date       : 10/05/09
Synopsis   : Calculate solvent accessible surface area of a molecule.
Description: 
Bugs       : 
See also   : 
"""

from pro_act2 import *

class SolventAccessibleSurfaceArea(ProAct):
    """
    This calculates the solvent accessible surface area of a molecule.
    """
    
    def _read_command_line(self):
        """
        Get command line options and arguements.

        """

        usage = "%prog [options] file1 [file2]"
        version = "%prog 0.1"
        parser = optparse.OptionParser(usage=usage, version=version)
        parser.add_option('--radii',
                          dest='radii_set',
                          default='LI_AND_NUSSINOV',
                          help='radii set used; WILLIAMS or LI_AND_NUSSINOV (defualt LI_AND_NUSSINOV)')
        parser.add_option('--res_sel1',
                          dest='res_sel1',
                          default='all',
                          help='residue selection for first file; e.g. 1,3,6:9')
        parser.add_option('--res_sel2',
                          dest='res_sel2',
                          default='all',
                          help='residue selection for second file; e.g. 1,3,6:9')
        parser.add_option('--chain_sel1',
                          dest='chain_sel1',
                          default='all',
                          help='chain selection for first file')
        parser.add_option('--chain_sel2',
                          dest='chain_sel2',
                          default='all',
                          help='chain selection for second file')
        parser.add_option('--file_type1',
                          dest='file_type1',
                          help='explicityly set the file type of first file (pdb or mol2)')
        parser.add_option('--file_type2',
                          dest='file_type2',
                          help='explicityly set the file type of second file (pdb or mol2)')
       
        options, args = parser.parse_args(sys.argv[1:])

        if len(args) < 1 or len(args) > 2:
            parser.error('Incorrect number of arguments')

        known_file_types = ['pdb', 'mol2']
        if not options.file_type1:
            options.file_type1 = args[0].split('.')[-1].strip()
        if not options.file_type2 and len(args) == 2:
            options.file_type2 = args[1].split('.')[-1].strip()
        if options.file_type1 not in known_file_types:
            parser.error('Unknown file type: %s' % options.file_type1)
        if options.file_type2 not in known_file_types and len(args) == 2:
            parser.error('Unknown file type: %s' % options.file_type2)

        if options.radii_set not in ['LI_AND_NUSSINOV', 'WILLIAMS']:
            parser.error("unrecognized argument for option radii_set")
        R.radii_set = options.radii_set

            

        options.chain_sel1 = self._chain_sel_to_list(options.chain_sel1)
        options.chain_sel2 = self._chain_sel_to_list(options.chain_sel2)

        options.res_sel1 = self._res_sel_to_list(options.res_sel1)
        options.res_sel2 = self._res_sel_to_list(options.res_sel2)

        options.experimental_waters = False

        if len(args) == 1:
            args.append(None)
            options.file_type2 = None
            options.chain_sel2 = None
            options.res_sel2 = None

        info = """
        ------------------------------------------------------
                   Structure File 1
        File                       : %s
        Type                       : %s
        Chain Selection            : %s
        Residue Selection          : %s
        ------------------------------------------------------
                   Structure File 2
        File                       : %s
        Type                       : %s
        Chain Selection            : %s
        Residue Selection          : %s
        ------------------------------------------------------
                            Options
        Radii Set                  : %s
        ------------------------------------------------------
            """ % (args[0],
                   options.file_type1,
                   options.chain_sel1,
                   options.res_sel1,
                   args[1],
                   options.file_type2,
                   options.chain_sel2,
                   options.res_sel2,
                   options.radii_set)
        
        self.args = args
        self.options = options
        self.info = re.sub(r'\n\s+', '\n',  info)

    def main(self):
        """
        This is the function that executes the rest of the program.
        """

        if __debug__:
            log = open('surface_area.log', 'w')

        # Read the command line
        self._read_command_line()
        if __debug__:
            log.write(self.info)
        
        # Read in the first file
        if self.options.file_type1 == 'pdb':
            M, H = self._read_pdb(molecule_instance=1,
                           file_name=self.args[0],
                           residue_selection=self.options.res_sel1,
                           chain_selection=self.options.chain_sel1)
            self.A.macro_molecule.extend(M)
            self.A.hydration.extend(H)
        elif self.options.file_type1 == 'mol2':
            mol2 = self._read_mol2(molecule_instance=1,
                             file_name=self.args[0])
            self.A.ligand.extend(mol2)
        else:
            raise ProAct2Error('Unknown file type')

        # Read in the second file
        if self.args[1]:
            if self.options.file_type2 == 'pdb':
                M, H = self._read_pdb(molecule_instance=2,
                               file_name=self.args[1],
                               residue_selection=self.options.res_sel2,
                               chain_selection=self.options.chain_sel2)
                self.A.macro_molecule.extend(M)
                self.A.hydration.extend(H)
            elif self.options.file_type2 == 'mol2':
                mol2 = self._read_mol2(molecule_instance=2,
                                 file_name=self.args[1])
                self.A.ligand.extend(mol2)
            else:
                raise ProAct2Error('Unknown file type')

        # Use first half of file1 as output file prefix
        first_file_name = os.path.split(self.args[0])
        file_prefix = first_file_name[1].split('.')[0]

        # Select all the protein and ligand atoms
        atoms = self.A.macro_molecule + self.A.ligand
        Assembly.set_neighbors(atoms)
        for atom in atoms:
            atom.set_surface_points(self.U, atom.close_radius, 0, 0)
            atom.set_accessibility(self.U)

        # Print the output
        prefix = '%s_' % file_prefix
        acc_file_name = prefix + "accessibility.kin"
        acc_file = open(acc_file_name, mode="w")
        acc_text = self.A.get_kin_accessibility(atoms)
        acc_file.write(acc_text)
        acc_file.close()


        

if __name__ == '__main__':
    S = SolventAccessibleSurfaceArea()
    S.main()

